# This example demonstrates how to add non-standard residues to
# sequences handled by the Python IHM library. See the simple-docking.py
# example for an introduction to the library.

import ihm
import ihm.dumper

system = ihm.System()

# IHM contains definitions for standard amino and nucleic acids, plus
# a few common non-standard residues such as MSE and UNK.
# To create a new non-standard residue, we first need to create a chemical
# component for it. In this case, we add a definition for norvaline, an
# isomer of valine. IHM provides a ChemComp class for this purpose and a
# number of more specialized subclasses. Since norvaline is a chiral peptide,
# here we define it in its L- form using the LPeptideChemComp class.
#
# 'id' should match the officially defined name of the component, as defined
#     in the chemical component dictionary: https://www.wwpdb.org/data/ccd
#     (See also https://www3.rcsb.org/ligand/NVA)
# 'code' is used to populate the primary sequence in the output mmCIF file.
#     For non-standard residues it should normally match 'id'.
# 'code_canonical' is the one-letter code of the closest standard residue.
#     Here we use 'V', valine.
norvaline = ihm.LPeptideChemComp(id='NVA', code='NVA', code_canonical='V',
                                 name='NORVALINE', formula='C5 H11 N O2')

# The Entity constructor takes a sequence of either or both one-letter codes
# and ChemComp objects, so now we can make a sequence containing both
# alanine and norvaline:
entity1 = ihm.Entity(['A', 'A', norvaline, 'A'], description='First entity')


# If a particular non-standard residue is commonly used in your own software,
# and you have assigned a one-letter code for it, you can subclass
# the ihm Alphabet class appropriately. Here we extend the normal set of
# one-letter codes (uppercase) for standard L- amino acids to add 'n' for
# norvaline:
class MyAlphabet(ihm.LPeptideAlphabet):
    # Alphabet contains a _comps dictionary that is a simple mapping from
    # codes (usually one-letter) to ChemComp objects
    _comps = {}
    _comps.update(ihm.LPeptideAlphabet._comps)
    _comps['n'] = norvaline


# Now we can pass a primary sequence using our custom alphabet to include
# norvaline alongside standard cysteine:
entity2 = ihm.Entity('CCnC', alphabet=MyAlphabet, description="Second entity")

system.entities.extend((entity1, entity2))

# Once the system is complete, we can write it out to an mmCIF file:
with open('output.cif', 'w') as fh:
    ihm.dumper.write(fh, [system])



# This example demonstrates the use of the Python IHM library at a low
# level, to parse an mmCIF file and extract a subset of its data.
# This particular example just extracts the atomic coordinates.

import ihm.format


# Make an object to handle a given mmCIF category in the file; it will
# be called for each line in the loop construct.
class AtomSiteHandler(object):
    # If a given keyword is not in the file, or has the special
    # mmCIF omitted (.) or unknown (?) value, the corresponding argument
    # to __call__ will be given these values:
    not_in_file = omitted = None
    unknown = ihm.unknown

    # Extract the group_PDB, Cartn_x, Cartn_y, Cartn_z keywords from
    # the mmCIF category (mmCIF keywords are case-insensitive, but the
    # Python arguments here should be lowercase).
    def __call__(self, group_pdb, cartn_x, cartn_y, cartn_z):
        if group_pdb == 'ATOM':
            print("Atom at %s, %s, %s" % (cartn_x, cartn_y, cartn_z))


ash = AtomSiteHandler()
with open('mini.cif') as fh:
    # Extract keywords from the _atom_site mmCIF category using the
    # AtomSiteHandler defined above
    c = ihm.format.CifReader(fh, category_handler={'_atom_site': ash})

    # Read the first data block in mini.cif
    # (This will return True as long as there are more blocks, so it can
    # be put in a while loop.)
    c.read_file()



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.flr


class Tests(unittest.TestCase):

    def test_Probe_Init(self):
        """Test initialization of probe_list_entry and probe_descriptor."""
        p = ihm.flr.Probe(probe_list_entry='foo', probe_descriptor='bar')
        self.assertEqual(p.probe_list_entry, 'foo')
        self.assertEqual(p.probe_descriptor, 'bar')

    def test_Probe_Eq(self):
        """Test equality and inequality of Probe objects"""
        p_ref = ihm.flr.Probe(probe_list_entry='foo', probe_descriptor='bar')
        p_equal = ihm.flr.Probe(probe_list_entry='foo', probe_descriptor='bar')
        p_unequal = ihm.flr.Probe(
            probe_list_entry='foo2', probe_descriptor='bar')
        self.assertTrue(p_ref == p_equal)
        self.assertFalse(p_ref == p_unequal)
        self.assertTrue(p_ref != p_unequal)

    def test_probe_descriptor_init(self):
        """ Test initalization of ProbeDescriptor """
        p = ihm.flr.ProbeDescriptor(reactive_probe_chem_descriptor='foo',
                                    chromophore_chem_descriptor='bar',
                                    chromophore_center_atom='foo2')
        self.assertEqual(p.reactive_probe_chem_descriptor, 'foo')
        self.assertEqual(p.chromophore_chem_descriptor, 'bar')
        self.assertEqual(p.chromophore_center_atom, 'foo2')

    def test_probe_descriptor_eq(self):
        """Test equality and inequality of ProbeDescriptor objects."""
        p_ref = ihm.flr.ProbeDescriptor(reactive_probe_chem_descriptor='foo',
                                        chromophore_chem_descriptor='bar',
                                        chromophore_center_atom='foo2')
        p_equal = ihm.flr.ProbeDescriptor(reactive_probe_chem_descriptor='foo',
                                          chromophore_chem_descriptor='bar',
                                          chromophore_center_atom='foo2')
        p_unequal = ihm.flr.ProbeDescriptor(
            reactive_probe_chem_descriptor='foo',
            chromophore_chem_descriptor='bar2', chromophore_center_atom='foo2')
        self.assertTrue(p_ref == p_equal)
        self.assertFalse(p_ref == p_unequal)
        self.assertTrue(p_ref != p_unequal)

    def test_probe_list_init(self):
        """ Test initialization of ProbeList. """
        p = ihm.flr.ProbeList(chromophore_name='foo',
                              reactive_probe_flag=False,
                              reactive_probe_name='bar',
                              probe_origin='foo2',
                              probe_link_type='bar2')
        self.assertEqual(p.chromophore_name, 'foo')
        self.assertEqual(p.reactive_probe_flag, False)
        self.assertEqual(p.reactive_probe_name, 'bar')
        self.assertEqual(p.probe_origin, 'foo2')
        self.assertEqual(p.probe_link_type, 'bar2')

    def test_probe_list_eq(self):
        """Test equality and inequality of ProbeList objects."""
        p_ref = ihm.flr.ProbeList(chromophore_name='foo',
                                  reactive_probe_flag=False,
                                  reactive_probe_name='bar',
                                  probe_origin='foo2',
                                  probe_link_type='bar2')
        p_equal = ihm.flr.ProbeList(chromophore_name='foo',
                                    reactive_probe_flag=False,
                                    reactive_probe_name='bar',
                                    probe_origin='foo2',
                                    probe_link_type='bar2')
        p_unequal = ihm.flr.ProbeList(chromophore_name='foo2',
                                      reactive_probe_flag=True,
                                      reactive_probe_name='bar',
                                      probe_origin='foo2',
                                      probe_link_type='bar2')
        self.assertTrue(p_ref == p_equal)
        self.assertFalse(p_ref == p_unequal)
        self.assertTrue(p_ref != p_unequal)

    def test_sample_probe_details_init(self):
        """ Test initialization of SampleProbeDetails. """
        s = ihm.flr.SampleProbeDetails(sample='foo',
                                       probe='bar',
                                       description='foo2',
                                       poly_probe_position='bar2')
        self.assertEqual(s.sample, 'foo')
        self.assertEqual(s.probe, 'bar')
        self.assertEqual(s.fluorophore_type, 'unspecified')
        self.assertEqual(s.description, 'foo2')
        self.assertEqual(s.poly_probe_position, 'bar2')

    def test_sample_probe_details_eq(self):
        """Test equality and inequality of SampleProbeDetails objects."""
        s_ref = ihm.flr.SampleProbeDetails(
            sample='foo', probe='bar', description='foo2',
            poly_probe_position='bar2')
        s_equal = ihm.flr.SampleProbeDetails(
            sample='foo', probe='bar', description='foo2',
            poly_probe_position='bar2')
        s_unequal = ihm.flr.SampleProbeDetails(
            sample='foo', probe='bar3', description='foo2',
            poly_probe_position='bar2')

        self.assertTrue(s_ref == s_equal)
        self.assertFalse(s_ref == s_unequal)
        self.assertTrue(s_ref != s_unequal)

    def test_poly_probe_conjugate_init(self):
        """ Test initialization of PolyProbeConjugate. """
        p = ihm.flr.PolyProbeConjugate(sample_probe='foo',
                                       chem_descriptor='bar',
                                       ambiguous_stoichiometry=True,
                                       probe_stoichiometry=0.5)
        self.assertEqual(p.sample_probe, 'foo')
        self.assertEqual(p.chem_descriptor, 'bar')
        self.assertEqual(p.ambiguous_stoichiometry, True)
        self.assertEqual(p.probe_stoichiometry, 0.5)

    def test_poly_probe_conjugate_eq(self):
        """ Test equality and inequality of PolyProbeConjugate objects. """
        p_ref = ihm.flr.PolyProbeConjugate(sample_probe='foo',
                                           chem_descriptor='bar',
                                           ambiguous_stoichiometry=True,
                                           probe_stoichiometry=0.5)
        p_equal = ihm.flr.PolyProbeConjugate(sample_probe='foo',
                                             chem_descriptor='bar',
                                             ambiguous_stoichiometry=True,
                                             probe_stoichiometry=0.5)
        p_unequal = ihm.flr.PolyProbeConjugate(
            sample_probe='foo2', chem_descriptor='bar',
            ambiguous_stoichiometry=True, probe_stoichiometry=0.5)
        self.assertTrue(p_ref == p_equal)
        self.assertFalse(p_ref == p_unequal)
        self.assertTrue(p_ref != p_unequal)

    def test_poly_probe_position_init(self):
        """ Test initialization of PolyProbePosition. """
        p = ihm.flr.PolyProbePosition(resatom='foo',
                                      mutation_flag=True,
                                      modification_flag=True,
                                      auth_name='foo3',
                                      mutated_chem_comp_id='foobar',
                                      modified_chem_descriptor='foobar2')
        self.assertEqual(p.resatom, 'foo')
        self.assertEqual(p.mutation_flag, True)
        self.assertEqual(p.modification_flag, True)
        self.assertEqual(p.auth_name, 'foo3')
        self.assertEqual(p.mutated_chem_comp_id, 'foobar')
        self.assertEqual(p.modified_chem_descriptor, 'foobar2')

    def test_poly_probe_position_eq(self):
        """Test equality and inequality of PolyProbePosition objects."""
        p_ref = ihm.flr.PolyProbePosition(
            resatom='foo', mutation_flag=True, modification_flag=True,
            auth_name='foo3', mutated_chem_comp_id='foobar',
            modified_chem_descriptor='foobar2')
        p_equal = ihm.flr.PolyProbePosition(
            resatom='foo', mutation_flag=True, modification_flag=True,
            auth_name='foo3', mutated_chem_comp_id='foobar',
            modified_chem_descriptor='foobar2')
        p_unequal = ihm.flr.PolyProbePosition(
            resatom='bar', mutation_flag=True, modification_flag=True,
            auth_name='foo3', mutated_chem_comp_id='foobar',
            modified_chem_descriptor='foobar2')

        self.assertTrue(p_ref == p_equal)
        self.assertFalse(p_ref == p_unequal)
        self.assertTrue(p_ref != p_unequal)

    def test_sample_init(self):
        """Test initialization of Sample."""
        s = ihm.flr.Sample(entity_assembly='foo', num_of_probes='bar',
                           condition='foo2', description='foo3',
                           details='foo4', solvent_phase='foobar')
        self.assertEqual(s.entity_assembly, 'foo')
        self.assertEqual(s.num_of_probes, 'bar')
        self.assertEqual(s.condition, 'foo2')
        self.assertEqual(s.description, 'foo3')
        self.assertEqual(s.details, 'foo4')
        self.assertEqual(s.solvent_phase, 'foobar')

    def test_sample_eq(self):
        """Test equality and inequality of Sample objects."""
        s_ref = ihm.flr.Sample(entity_assembly='foo', num_of_probes='foo2',
                               condition='foo3', description='foo4',
                               details='foo5', solvent_phase='foo6')
        s_equal = ihm.flr.Sample(entity_assembly='foo', num_of_probes='foo2',
                                 condition='foo3', description='foo4',
                                 details='foo5', solvent_phase='foo6')
        s_unequal = ihm.flr.Sample(entity_assembly='bar', num_of_probes='bar2',
                                   condition='bar3', description='bar4',
                                   details='bar5', solvent_phase='bar6')
        self.assertTrue(s_ref == s_equal)
        self.assertFalse(s_ref == s_unequal)
        self.assertTrue(s_ref != s_unequal)

    def test_entity_assembly_init(self):
        """ Test initialization of EntityAssembly. """
        e = ihm.flr.EntityAssembly(entity='foo', num_copies=1)
        self.assertEqual(len(e.entity_list), 1)
        self.assertEqual(e.entity_list[0], 'foo')
        self.assertEqual(e.num_copies_list[0], 1)

    def test_entity_assembly_add_entity(self):
        """Test addition of entities to the assembly."""
        e = ihm.flr.EntityAssembly(entity='foo', num_copies=1)
        e.add_entity(entity='foo2', num_copies=2)
        self.assertRaises(ValueError, e.add_entity,
                          entity='foo2', num_copies=-1)
        self.assertEqual(len(e.entity_list), 2)
        self.assertEqual(e.entity_list, ['foo', 'foo2'])
        self.assertEqual(e.num_copies_list, [1, 2])

    def test_entity_assembly_eq(self):
        """ Test equality and inequality of EntityAssembly objects. """
        e_ref = ihm.flr.EntityAssembly(entity='foo', num_copies=1)
        e_equal = ihm.flr.EntityAssembly(entity='foo', num_copies=1)
        e_unequal = ihm.flr.EntityAssembly(entity='foo2', num_copies=1)
        self.assertTrue(e_ref == e_equal)
        self.assertFalse(e_ref == e_unequal)
        self.assertTrue(e_ref != e_unequal)

    def test_sample_condition_init(self):
        """ Test initialization of SampleCondition. """
        s = ihm.flr.SampleCondition(details='foo')
        self.assertEqual(s.details, 'foo')

    def test_sample_condition_eq(self):
        """ Test equality and inequality of SampleCondition objects. """
        s_ref = ihm.flr.SampleCondition(details='foo')
        s_equal = ihm.flr.SampleCondition(details='foo')
        s_unequal = ihm.flr.SampleCondition(details='bar')
        self.assertTrue(s_ref == s_equal)
        self.assertFalse(s_ref == s_unequal)
        self.assertTrue(s_ref != s_unequal)

    def test_Experiment_Init(self):
        """Test initialization of Experiment."""
        # Initialization with only one parameter given should not add an entry
        e1 = ihm.flr.Experiment(instrument='foo')
        self.assertEqual(len(e1.instrument_list), 0)
        self.assertEqual(len(e1.inst_setting_list), 0)
        self.assertEqual(len(e1.exp_condition_list), 0)
        self.assertEqual(len(e1.sample_list), 0)

        # Correct initialization should fill the lists
        e2 = ihm.flr.Experiment(instrument='foo',
                                inst_setting='bar',
                                exp_condition='foo2',
                                sample='foo3',
                                details='bar2')
        self.assertEqual(len(e2.instrument_list), 1)
        self.assertEqual(e2.instrument_list[0], 'foo')
        self.assertEqual(e2.inst_setting_list[0], 'bar')
        self.assertEqual(e2.exp_condition_list[0], 'foo2')
        self.assertEqual(e2.sample_list[0], 'foo3')
        self.assertEqual(e2.details_list[0], 'bar2')

        # Initialization without details given should still have an entry
        # in the list
        e3 = ihm.flr.Experiment(instrument='foo',
                                inst_setting='bar',
                                exp_condition='bar2',
                                sample='foo2')
        self.assertEqual(len(e3.details_list), 1)
        self.assertIsNone(e3.details_list[0])

    def test_Experiment_Add_entry(self):
        """ Test addition of an entry to the experiment. """
        # Adding to an empty Experiment
        e1 = ihm.flr.Experiment()
        e1.add_entry(instrument='foo',
                     inst_setting='bar',
                     exp_condition='foo2',
                     sample='foo3',
                     details='bar2')
        self.assertEqual(e1.instrument_list[0], 'foo')
        self.assertEqual(e1.inst_setting_list[0], 'bar')
        self.assertEqual(e1.exp_condition_list[0], 'foo2')
        self.assertEqual(e1.sample_list[0], 'foo3')
        self.assertEqual(e1.details_list[0], 'bar2')

        # adding to an existing Experiment
        e2 = ihm.flr.Experiment(instrument='foo',
                                inst_setting='foo2',
                                exp_condition='foo3',
                                sample='foo4',
                                details='foo5')
        e2.add_entry(instrument='bar',
                     inst_setting='bar2',
                     exp_condition='bar3',
                     sample='bar4',
                     details='bar5')
        self.assertEqual(e2.instrument_list, ['foo', 'bar'])
        self.assertEqual(e2.inst_setting_list, ['foo2', 'bar2'])
        self.assertEqual(e2.exp_condition_list, ['foo3', 'bar3'])
        self.assertEqual(e2.sample_list, ['foo4', 'bar4'])
        self.assertEqual(e2.details_list, ['foo5', 'bar5'])

    def test_Experiment_Get_entry_by_index(self):
        """ Test access to entries by index. """
        e = ihm.flr.Experiment()
        e.add_entry(instrument='foo', inst_setting='foo2',
                    exp_condition='foo3', sample='foo4', details='foo5')
        e.add_entry(instrument='bar', inst_setting='bar2',
                    exp_condition='bar3', sample='bar4', details='bar5')
        e.add_entry(instrument='foobar', inst_setting='foobar2',
                    exp_condition='foobar3', sample='foobar4',
                    details='foobar5')
        return_value_index0 = e.get_entry_by_index(0)
        return_value_index1 = e.get_entry_by_index(1)
        return_value_index2 = e.get_entry_by_index(2)
        self.assertEqual(return_value_index0,
                         ('foo', 'foo2', 'foo3', 'foo4', 'foo5'))
        self.assertEqual(return_value_index1,
                         ('bar', 'bar2', 'bar3', 'bar4', 'bar5'))
        self.assertEqual(return_value_index2,
                         ('foobar', 'foobar2', 'foobar3', 'foobar4',
                          'foobar5'))

    def test_Experiment_Contains(self):
        """Test whether experiment contains a combination of instrument,
           exp_setting, and sample."""
        # An empty experiment should not contain anything
        e1 = ihm.flr.Experiment()
        self.assertFalse(e1.contains('foo', 'foo2', 'foo3', 'foo4'))
        # After addition, the entry should be contained
        e1.add_entry(instrument='foo', inst_setting='foo2',
                     exp_condition='foo3', sample='foo4')
        e1.add_entry(instrument='bar', inst_setting='bar2',
                     exp_condition='bar3', sample='bar4')
        self.assertTrue(e1.contains('foo', 'foo2', 'foo3', 'foo4'))
        # If one of the entries is not contained, then False
        self.assertFalse(e1.contains('foo2', 'foo2', 'foo4', 'foo5'))
        self.assertFalse(e1.contains('foobar', 'foobar2', 'foobar3',
                                     'foobar4'))

    def test_Experiment_Eq(self):
        """ Test equality and inequality of Experiment objects. """
        e_ref = ihm.flr.Experiment()
        e_ref.add_entry(instrument='foo', inst_setting='foo2',
                        exp_condition='foo3', sample='foo4')
        e_equal = ihm.flr.Experiment()
        e_equal.add_entry(instrument='foo', inst_setting='foo2',
                          exp_condition='foo3', sample='foo4')
        e_unequal = ihm.flr.Experiment()
        e_unequal.add_entry(instrument='bar', inst_setting='bar2',
                            exp_condition='bar3', sample='bar4')
        self.assertTrue(e_ref == e_equal)
        self.assertFalse(e_ref == e_unequal)
        self.assertTrue(e_ref != e_unequal)

    def test_Instrument_Init(self):
        """ Test initialization of Instrument. """
        i = ihm.flr.Instrument(details='foo')
        self.assertEqual(i.details, 'foo')

    def test_Instrument_Eq(self):
        """Test equality and inequality of Instrument objects."""
        i_ref = ihm.flr.Instrument(details='foo')
        i_equal = ihm.flr.Instrument(details='foo')
        i_unequal = ihm.flr.Instrument(details='bar')
        self.assertTrue(i_ref == i_equal)
        self.assertFalse(i_ref == i_unequal)
        self.assertTrue(i_ref != i_unequal)

    def test_inst_setting_init(self):
        """Test initialization of InstSetting."""
        e = ihm.flr.InstSetting(details='foo')
        self.assertEqual(e.details, 'foo')

    def test_inst_setting_eq(self):
        """Test equality and inequality of InstSetting objects."""
        e_ref = ihm.flr.InstSetting(details='foo')
        e_equal = ihm.flr.InstSetting(details='foo')
        e_unequal = ihm.flr.InstSetting(details='bar')
        self.assertTrue(e_ref == e_equal)
        self.assertFalse(e_ref == e_unequal)
        self.assertTrue(e_ref != e_unequal)

    def test_exp_condition_init(self):
        """Test initialization of ExpCondition."""
        e = ihm.flr.ExpCondition(details='foo')
        self.assertEqual(e.details, 'foo')

    def test_exp_condition_eq(self):
        """Test equality and inequality of ExpCondition objects."""
        e_ref = ihm.flr.ExpCondition(details='foo')
        e_equal = ihm.flr.ExpCondition(details='foo')
        e_unequal = ihm.flr.ExpCondition(details='bar')
        self.assertTrue(e_ref == e_equal)
        self.assertFalse(e_ref == e_unequal)
        self.assertTrue(e_ref != e_unequal)

    def test_fret_analysis_init(self):
        """Test initalization of FRETAnalysis."""
        f = ihm.flr.FRETAnalysis(
            experiment='this_experiment',
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            forster_radius='this_forster_radius',
            type='intensity-based',
            calibration_parameters='this_calibration_parameters',
            lifetime_fit_model='this_lifetime_fit_model',
            ref_measurement_group='this_ref_measurement_group',
            method_name='this_method_name',
            chi_square_reduced='this_chi_square_reduced',
            donor_only_fraction='this_donly_fraction',
            dataset='this_dataset_list_id',
            external_file='this_external_file',
            software='this_software')
        self.assertEqual(f.experiment, 'this_experiment')
        self.assertEqual(f.sample_probe_1, 'this_sample_probe_1')
        self.assertEqual(f.sample_probe_2, 'this_sample_probe_2')
        self.assertEqual(f.forster_radius, 'this_forster_radius')
        self.assertEqual(f.type, 'intensity-based')
        self.assertEqual(f.calibration_parameters,
                         'this_calibration_parameters')
        self.assertEqual(f.lifetime_fit_model, 'this_lifetime_fit_model')
        self.assertEqual(f.ref_measurement_group, 'this_ref_measurement_group')
        self.assertEqual(f.method_name, 'this_method_name')
        self.assertEqual(f.chi_square_reduced, 'this_chi_square_reduced')
        self.assertEqual(f.donor_only_fraction, 'this_donly_fraction')
        self.assertEqual(f.dataset, 'this_dataset_list_id')
        self.assertEqual(f.external_file, 'this_external_file')
        self.assertEqual(f.software, 'this_software')

        self.assertRaises(
            ValueError,
            ihm.flr.FRETAnalysis, experiment='this_experiment',
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            forster_radius='this_forster_radius', type='garbage')

    def test_fret_analysis_eq(self):
        """Test equality and inequality of FRETAnalysis objects."""
        f_ref = ihm.flr.FRETAnalysis(
            experiment='this_experiment',
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            forster_radius='this_forster_radius',
            type='intensity-based',
            calibration_parameters='this_calibration_parameters',
            method_name='this_method_name',
            chi_square_reduced='this_chi_square_reduced',
            donor_only_fraction='this_donly_fraction',
            dataset='this_dataset_list_id',
            external_file='this_external_file',
            software='this_software')
        f_equal = ihm.flr.FRETAnalysis(
            experiment='this_experiment',
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            forster_radius='this_forster_radius',
            type='intensity-based',
            calibration_parameters='this_calibration_parameters',
            method_name='this_method_name',
            chi_square_reduced='this_chi_square_reduced',
            donor_only_fraction='this_donly_fraction',
            dataset='this_dataset_list_id',
            external_file='this_external_file',
            software='this_software')
        f_unequal = ihm.flr.FRETAnalysis(
            experiment='this_experiment',
            sample_probe_1='foo',
            sample_probe_2='this_sample_probe_2',
            forster_radius='this_forster_radius',
            type='intensity-based',
            calibration_parameters='this_calibration_parameters',
            method_name='this_method_name',
            chi_square_reduced='this_chi_square_reduced',
            donor_only_fraction='this_donly_fraction',
            dataset='this_dataset_list_id',
            external_file='this_external_file',
            software='this_software')
        f_unequal_type = ihm.flr.FRETAnalysis(
            experiment='this_experiment',
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            forster_radius='this_forster_radius',
            type='lifetime-based',
            calibration_parameters='this_calibration_parameters',
            method_name='this_method_name',
            chi_square_reduced='this_chi_square_reduced',
            donor_only_fraction='this_donly',
            dataset='this_dataset_list_id',
            external_file='this_external_file',
            software='this_software')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)
        self.assertFalse(f_ref == f_unequal_type)
        self.assertTrue(f_ref != f_unequal_type)

    def test_lifetime_fit_model_init(self):
        """ Test initialization of LifetimeFitModel."""
        f = ihm.flr.LifetimeFitModel(name='this_name',
                                     description='this_description',
                                     external_file='this_ext_file',
                                     citation='this_citation')
        self.assertEqual(f.name, 'this_name')
        self.assertEqual(f.description, 'this_description')
        self.assertEqual(f.external_file, 'this_ext_file')
        self.assertEqual(f.citation, 'this_citation')

    def test_lifetime_fit_model_eq(self):
        """ Test equality and inequality of LifeTimeFitModel objects."""
        f_ref = ihm.flr.LifetimeFitModel(
            name='this_name', description='this_desc')
        f_equal = ihm.flr.LifetimeFitModel(
            name='this_name', description='this_desc')
        f_unequal = ihm.flr.LifetimeFitModel(
            name='other_name', description='this_desc')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_ref_measurement_group_init(self):
        """ Test initialization of RefMeasurementGroup."""
        r = ihm.flr.RefMeasurementGroup()
        self.assertEqual(r.ref_measurement_list, [])

    def test_ref_measurement_group_add_ref_measurement(self):
        """ Test the addition of a RefMeasurement to the group."""
        r = ihm.flr.RefMeasurementGroup()
        r.add_ref_measurement('foo')
        r.add_ref_measurement('bar')
        self.assertEqual(r.ref_measurement_list, ['foo', 'bar'])

    def test_ref_measurement_group_get_info(self):
        """ Test the retrieval of the ref_measurement_list."""
        r = ihm.flr.RefMeasurementGroup()
        r.add_ref_measurement('foo')
        r.add_ref_measurement('bar')
        return_value = r.get_info()
        self.assertEqual(return_value, ['foo', 'bar'])

    def test_ref_measurement_group_eq(self):
        """ Test equality and inequality of RefMeasurementGroup objects."""
        r_ref = ihm.flr.RefMeasurementGroup()
        r_ref.add_ref_measurement('foo')
        r_equal = ihm.flr.RefMeasurementGroup()
        r_equal.add_ref_measurement('foo')
        r_unequal = ihm.flr.RefMeasurementGroup()
        r_unequal.add_ref_measurement('foo2')
        self.assertTrue(r_ref == r_equal)
        self.assertFalse(r_ref == r_unequal)
        self.assertTrue(r_ref != r_unequal)

    def test_ref_measurement_init(self):
        """Test initialization of RefMeasurement."""
        r1 = ihm.flr.RefMeasurement(ref_sample_probe='this_ref_sample_probe',
                                    details='this_details')
        self.assertEqual(r1.ref_sample_probe, 'this_ref_sample_probe')
        self.assertEqual(r1.details, 'this_details')
        self.assertEqual(r1.list_of_lifetimes, [])

        r2 = ihm.flr.RefMeasurement(ref_sample_probe='this_ref_sample_probe_2',
                                    details='this_details_2',
                                    list_of_lifetimes=['foo', 'bar'])
        self.assertEqual(r2.ref_sample_probe, 'this_ref_sample_probe_2')
        self.assertEqual(r2.details, 'this_details_2')
        self.assertEqual(r2.list_of_lifetimes, ['foo', 'bar'])

    def test_ref_measurement_add_lifetime(self):
        """ Test addition of to the list_of_lifetimes."""
        r = ihm.flr.RefMeasurement(ref_sample_probe='this_ref_sample_probe',
                                   details='this_details')
        r.add_lifetime('foo')
        r.add_lifetime('bar')
        self.assertEqual(r.list_of_lifetimes, ['foo', 'bar'])

    def test_ref_measurement_eq(self):
        """ Test equality and inequality of RefMeasurement objects."""
        r_ref = ihm.flr.RefMeasurement(
            ref_sample_probe='this_ref_sample_probe_1',
            details='this_details_1')
        r_equal = ihm.flr.RefMeasurement(
            ref_sample_probe='this_ref_sample_probe_1',
            details='this_details_1')
        r_unequal = ihm.flr.RefMeasurement(
            ref_sample_probe='this_ref_sample_probe_2',
            details='this_details_2')
        r_unequal_list = ihm.flr.RefMeasurement(
            ref_sample_probe='this_ref_sample_probe_1',
            details='this_details_1', list_of_lifetimes=['foo'])
        self.assertTrue(r_ref == r_equal)
        self.assertTrue(r_ref != r_unequal)
        self.assertTrue(r_ref != r_unequal_list)
        self.assertFalse(r_ref == r_unequal)
        self.assertFalse(r_ref == r_unequal_list)

    def test_ref_measurement_lifetime_init(self):
        """ Test initialization of RefMeasuremenLifetime objects."""
        f = ihm.flr.RefMeasurementLifetime(species_fraction='this_frac',
                                           lifetime='this_lifetime',
                                           species_name='foo')
        self.assertEqual(f.species_fraction, 'this_frac')
        self.assertEqual(f.lifetime, 'this_lifetime')
        self.assertEqual(f.species_name, 'foo')

    def test_ref_measurement_lifetime_eq(self):
        """ Test equality and inequality of RefMeasurementLifetime objects."""
        f_ref = ihm.flr.RefMeasurementLifetime(species_fraction='this_frac_1',
                                               lifetime='this_lifetime_1')
        f_equal = ihm.flr.RefMeasurementLifetime(
            species_fraction='this_frac_1', lifetime='this_lifetime_1')
        f_unequal = ihm.flr.RefMeasurementLifetime(
            species_fraction='this_frac_2', lifetime='this_lifetime_1')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fret_distance_restraint_group_init(self):
        """ Test initialization of FRETDistanceRestraintGroup. """
        f = ihm.flr.FRETDistanceRestraintGroup()
        self.assertEqual(f.distance_restraint_list, [])

    def test_fret_distance_restraint_group_add_distance_restraint(self):
        """ Test the addition of a Fret_distance restraint to the group. """
        f = ihm.flr.FRETDistanceRestraintGroup()
        f.add_distance_restraint('foo')
        f.add_distance_restraint('bar')
        self.assertEqual(f.distance_restraint_list, ['foo', 'bar'])

    def test_fret_distance_restraint_group_get_info(self):
        """ Test the retrieval of the distance_restraint_list. """
        f = ihm.flr.FRETDistanceRestraintGroup()
        f.add_distance_restraint('foo')
        f.add_distance_restraint('bar')
        return_value = f.get_info()
        self.assertEqual(return_value, ['foo', 'bar'])

    def test_fret_distance_restraint_group_eq(self):
        """Test equality and inequality of FRETDistanceRestraintGroup
           objects."""
        f_ref = ihm.flr.FRETDistanceRestraintGroup()
        f_ref.add_distance_restraint('foo')
        f_equal = ihm.flr.FRETDistanceRestraintGroup()
        f_equal.add_distance_restraint('foo')
        f_unequal = ihm.flr.FRETDistanceRestraintGroup()
        f_unequal.add_distance_restraint('bar')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fret_distance_restraint_init(self):
        """Test initialization of FRETDistanceRestraint."""
        f = ihm.flr.FRETDistanceRestraint(
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            state='this_state',
            analysis='this_analysis',
            distance='this_distance',
            distance_error_plus='this_distance_error_plus',
            distance_error_minus='this_distance_error_minus',
            distance_type='this_distance_type',
            population_fraction='this_population_fraction',
            peak_assignment='this_peak_assignment')
        self.assertEqual(f.sample_probe_1, 'this_sample_probe_1')
        self.assertEqual(f.sample_probe_2, 'this_sample_probe_2')
        self.assertEqual(f.state, 'this_state')
        self.assertEqual(f.analysis, 'this_analysis')
        self.assertEqual(f.distance, 'this_distance')
        self.assertEqual(f.distance_error_plus, 'this_distance_error_plus')
        self.assertEqual(f.distance_error_minus, 'this_distance_error_minus')
        self.assertEqual(f.distance_type, 'this_distance_type')
        self.assertEqual(f.population_fraction, 'this_population_fraction')
        self.assertEqual(f.peak_assignment, 'this_peak_assignment')

    def test_fret_distance_restraint_eq(self):
        """Test equality and inequality of FRETDistanceRestraint objects."""
        f_ref = ihm.flr.FRETDistanceRestraint(
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            state='this_state', analysis='this_analysis',
            distance='this_distance',
            distance_error_plus='this_distance_error_plus',
            distance_error_minus='this_distance_error_minus',
            distance_type='this_distance_type',
            population_fraction='this_population_fraction',
            peak_assignment='this_peak_assignment')

        f_equal = ihm.flr.FRETDistanceRestraint(
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2',
            state='this_state', analysis='this_analysis',
            distance='this_distance',
            distance_error_plus='this_distance_error_plus',
            distance_error_minus='this_distance_error_minus',
            distance_type='this_distance_type',
            population_fraction='this_population_fraction',
            peak_assignment='this_peak_assignment')

        f_unequal = ihm.flr.FRETDistanceRestraint(
            sample_probe_1='this_sample_probe_1',
            sample_probe_2='this_sample_probe_2', state='foo',
            analysis='this_analysis', distance='this_distance',
            distance_error_plus='this_distance_error_plus',
            distance_error_minus='this_distance_error_minus',
            distance_type='this_distance_type',
            population_fraction='this_population_fraction',
            peak_assignment='this_peak_assignment')

        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fret_forster_radius_init(self):
        """ Test initialization of FRETForsterRadius. """
        f = ihm.flr.FRETForsterRadius(
            donor_probe='foo', acceptor_probe='bar', forster_radius='foo2',
            reduced_forster_radius='bar2')
        self.assertEqual(f.donor_probe, 'foo')
        self.assertEqual(f.acceptor_probe, 'bar')
        self.assertEqual(f.forster_radius, 'foo2')
        self.assertEqual(f.reduced_forster_radius, 'bar2')

    def test_fret_forster_radius_eq(self):
        """ Test equality and inequality of FRETForsterRadius objects. """
        f_ref = ihm.flr.FRETForsterRadius(
            donor_probe='foo', acceptor_probe='bar', forster_radius='foo2',
            reduced_forster_radius='bar2')
        f_equal = ihm.flr.FRETForsterRadius(
            donor_probe='foo', acceptor_probe='bar', forster_radius='foo2',
            reduced_forster_radius='bar2')
        f_unequal = ihm.flr.FRETForsterRadius(
            donor_probe='foobar', acceptor_probe='bar', forster_radius='foo2',
            reduced_forster_radius='bar2')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fret_calibration_parameters_init(self):
        """Test initialization of FRETCalibrationParameters."""
        f = ihm.flr.FRETCalibrationParameters(phi_acceptor='this_phi_acceptor',
                                              alpha='this_alpha',
                                              alpha_sd='this_alpha_sd',
                                              gg_gr_ratio='this_gG_gR_ratio',
                                              beta='this_beta',
                                              gamma='this_gamma',
                                              delta='this_delta',
                                              a_b='this_a_b')
        self.assertEqual(f.phi_acceptor, 'this_phi_acceptor')
        self.assertEqual(f.alpha, 'this_alpha')
        self.assertEqual(f.alpha_sd, 'this_alpha_sd')
        self.assertEqual(f.gg_gr_ratio, 'this_gG_gR_ratio')
        self.assertEqual(f.beta, 'this_beta')
        self.assertEqual(f.gamma, 'this_gamma')
        self.assertEqual(f.delta, 'this_delta')
        self.assertEqual(f.a_b, 'this_a_b')

    def test_fret_calibration_parameters_eq(self):
        """Test equality and inequality of
           FRETCalibrationParameters objects."""
        f_ref = ihm.flr.FRETCalibrationParameters(
            phi_acceptor='this_phi_acceptor', alpha='this_alpha',
            alpha_sd='this_alpha_sd', gg_gr_ratio='this_gG_gR_ratio',
            beta='this_beta', gamma='this_gamma', delta='this_delta',
            a_b='this_a_b')
        f_equal = ihm.flr.FRETCalibrationParameters(
            phi_acceptor='this_phi_acceptor', alpha='this_alpha',
            alpha_sd='this_alpha_sd', gg_gr_ratio='this_gG_gR_ratio',
            beta='this_beta', gamma='this_gamma', delta='this_delta',
            a_b='this_a_b')
        f_unequal = ihm.flr.FRETCalibrationParameters(
            phi_acceptor='foo', alpha='this_alpha', alpha_sd='this_alpha_sd',
            gg_gr_ratio='this_gG_gR_ratio', beta='this_beta',
            gamma='this_gamma', delta='this_delta', a_b='this_a_b')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_peak_assignment_init(self):
        """ Test initialization of PeakAssignment. """
        p = ihm.flr.PeakAssignment(method_name='foo', details='bar')
        self.assertEqual(p.method_name, 'foo')
        self.assertEqual(p.details, 'bar')

    def test_peak_assignment_eq(self):
        """ Test equality and inequality of PeakAssignment objects. """
        p_ref = ihm.flr.PeakAssignment(method_name='foo', details='bar')
        p_equal = ihm.flr.PeakAssignment(method_name='foo', details='bar')
        p_unequal = ihm.flr.PeakAssignment(method_name='foobar', details='bar')
        self.assertTrue(p_ref == p_equal)
        self.assertFalse(p_ref == p_unequal)
        self.assertTrue(p_ref != p_unequal)

    def test_fret_model_quality_init(self):
        """ Test initialization of FRETModelQuality. """
        f = ihm.flr.FRETModelQuality(
            model='this_model_id',
            chi_square_reduced='this_chi_square_reduced',
            dataset_group='this_dataset_group_id', method='this_method',
            details='this_details')
        self.assertEqual(f.model, 'this_model_id')
        self.assertEqual(f.chi_square_reduced, 'this_chi_square_reduced')
        self.assertEqual(f.dataset_group, 'this_dataset_group_id')
        self.assertEqual(f.method, 'this_method')
        self.assertEqual(f.details, 'this_details')

    def test_fret_model_quality_eq(self):
        """ Test equality and inequality of FRETModelQuality objects. """
        f_ref = ihm.flr.FRETModelQuality(
            model='this_model_id',
            chi_square_reduced='this_chi_square_reduced',
            dataset_group='this_dataset_group_id', method='this_method',
            details='this_details')
        f_equal = ihm.flr.FRETModelQuality(
            model='this_model_id',
            chi_square_reduced='this_chi_square_reduced',
            dataset_group='this_dataset_group_id', method='this_method',
            details='this_details')
        f_unequal = ihm.flr.FRETModelQuality(
            model='foo', chi_square_reduced='this_chi_square_reduced',
            dataset_group='this_dataset_group_id', method='this_method',
            details='this_details')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fret_model_distance_init(self):
        """Test initialization of FRETModelDistance.
           Also test the implicit calculation of the distance deviation"""
        # Initialization with explicit setting of distance deviation
        f1 = ihm.flr.FRETModelDistance(restraint='foo', model='bar',
                                       distance=50, distance_deviation=4.0)
        self.assertEqual(f1.restraint, 'foo')
        self.assertEqual(f1.model, 'bar')
        self.assertEqual(f1.distance, 50)
        self.assertEqual(f1.distance_deviation, 4.0)

        # Initialization with calculation of distance deviation
        class Dummy_Restraint():
            def __init__(self, distance):
                self.distance = distance
        f2 = ihm.flr.FRETModelDistance(restraint=Dummy_Restraint(40),
                                       model='bar2', distance=30)
        self.assertEqual(f2.model, 'bar2')
        self.assertEqual(f2.distance, 30)
        self.assertEqual(f2.distance_deviation, 10.0)

    def test_fret_model_distance_calculate_deviation(self):
        """Test FRETModelDistance.calculate_deviation()"""
        class Dummy_Restraint():
            def __init__(self, distance):
                self.distance = distance
        f1 = ihm.flr.FRETModelDistance(restraint=Dummy_Restraint(40),
                                       model='foo', distance=30)
        self.assertEqual(f1.distance_deviation, 10.0)
        # Directly changing the distance should not change the deviation
        f1.distance = 25
        self.assertEqual(f1.distance_deviation, 10.0)
        # Calculation of the distance deviation should update the deviation
        f1.update_deviation()
        self.assertEqual(f1.distance_deviation, 15.0)
        # update_deviation() is a noop if restraint is None
        f2 = ihm.flr.FRETModelDistance(restraint=None, model='foo',
                                       distance=30)
        f2.update_deviation()
        self.assertIsNone(f2.distance_deviation)

    def test_fret_model_distance_eq(self):
        """ Test equality and inequality of FRETModelDistance objects. """
        f_ref = ihm.flr.FRETModelDistance(restraint='foo', model='bar',
                                          distance=50,
                                          distance_deviation=4.0)
        f_equal = ihm.flr.FRETModelDistance(restraint='foo', model='bar',
                                            distance=50,
                                            distance_deviation=4.0)
        f_unequal = ihm.flr.FRETModelDistance(restraint='foo2',
                                              model='bar', distance=50,
                                              distance_deviation=4.0)
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_modeling_init(self):
        """ Test initialization of FPSModeling. """
        f = ihm.flr.FPSModeling(protocol='foo',
                                restraint_group='bar',
                                global_parameter='foo2',
                                probe_modeling_method='foo3',
                                details='bar2')
        self.assertEqual(f.protocol, 'foo')
        self.assertEqual(f.restraint_group, 'bar')
        self.assertEqual(f.global_parameter, 'foo2')
        self.assertEqual(f.probe_modeling_method, 'foo3')
        self.assertEqual(f.details, 'bar2')

    def test_fps_modeling_eq(self):
        """Test equality and inequality of FPSModeling objects."""
        f_ref = ihm.flr.FPSModeling(protocol='foo', restraint_group='bar',
                                    global_parameter='foo2',
                                    probe_modeling_method='foo3',
                                    details='bar2')
        f_equal = ihm.flr.FPSModeling(protocol='foo', restraint_group='bar',
                                      global_parameter='foo2',
                                      probe_modeling_method='foo3',
                                      details='bar2')
        f_unequal = ihm.flr.FPSModeling(protocol='foo',
                                        restraint_group='foobar',
                                        global_parameter='foo2',
                                        probe_modeling_method='foo3',
                                        details='bar2')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_global_parameters_init(self):
        """Test initialization of FPSGlobalParameters."""
        f = ihm.flr.FPSGlobalParameters(
            forster_radius='this_forster_radius',
            conversion_function_polynom_order='this_cfpo',
            repetition='this_repetition',
            av_grid_rel='this_AV_grid_rel',
            av_min_grid_a='this_AV_min_grid_A',
            av_allowed_sphere='this_AV_allowed_sphere',
            av_search_nodes='this_AV_search_nodes',
            av_e_samples_k='this_AV_E_samples_k',
            sim_viscosity_adjustment='this_sim_viscosity_adjustment',
            sim_dt_adjustment='this_sim_dt_adjustment',
            sim_max_iter_k='this_sim_max_iter_k',
            sim_max_force='this_sim_max_force',
            sim_clash_tolerance_a='this_sim_clash_tolerance_A',
            sim_reciprocal_kt='this_sim_reciprocal_kT',
            sim_clash_potential='this_sim_clash_potential',
            convergence_e='this_convergence_E',
            convergence_k='this_convergence_K',
            convergence_f='this_convergence_F',
            convergence_t='this_convergence_T',
            optimized_distances='this_optimized_distances')

        self.assertEqual(f.forster_radius, 'this_forster_radius')
        self.assertEqual(f.conversion_function_polynom_order, 'this_cfpo')
        self.assertEqual(f.repetition, 'this_repetition')
        self.assertEqual(f.av_grid_rel, 'this_AV_grid_rel')
        self.assertEqual(f.av_min_grid_a, 'this_AV_min_grid_A')
        self.assertEqual(f.av_allowed_sphere, 'this_AV_allowed_sphere')
        self.assertEqual(f.av_search_nodes, 'this_AV_search_nodes')
        self.assertEqual(f.av_e_samples_k, 'this_AV_E_samples_k')
        self.assertEqual(f.sim_viscosity_adjustment,
                         'this_sim_viscosity_adjustment')
        self.assertEqual(f.sim_dt_adjustment, 'this_sim_dt_adjustment')
        self.assertEqual(f.sim_max_iter_k, 'this_sim_max_iter_k')
        self.assertEqual(f.sim_max_force, 'this_sim_max_force')
        self.assertEqual(f.sim_clash_tolerance_a, 'this_sim_clash_tolerance_A')
        self.assertEqual(f.sim_reciprocal_kt, 'this_sim_reciprocal_kT')
        self.assertEqual(f.sim_clash_potential, 'this_sim_clash_potential')
        self.assertEqual(f.convergence_e, 'this_convergence_E')
        self.assertEqual(f.convergence_k, 'this_convergence_K')
        self.assertEqual(f.convergence_f, 'this_convergence_F')
        self.assertEqual(f.convergence_t, 'this_convergence_T')
        self.assertEqual(f.optimized_distances, 'this_optimized_distances')

    def test_fps_global_parameters_eq(self):
        """Test equality and inequality of FPSGlobalParameters objects."""
        f_ref = ihm.flr.FPSGlobalParameters(
            forster_radius='this_forster_radius',
            conversion_function_polynom_order='this_cfpo',
            repetition='this_repetition',
            av_grid_rel='this_AV_grid_rel',
            av_min_grid_a='this_AV_min_grid_A',
            av_allowed_sphere='this_AV_allowed_sphere',
            av_search_nodes='this_AV_search_nodes',
            av_e_samples_k='this_AV_E_samples_k',
            sim_viscosity_adjustment='this_sim_viscosity_adjustment',
            sim_dt_adjustment='this_sim_dt_adjustment',
            sim_max_iter_k='this_sim_max_iter_k',
            sim_max_force='this_sim_max_force',
            sim_clash_tolerance_a='this_sim_clash_tolerance_A',
            sim_reciprocal_kt='this_sim_reciprocal_kT',
            sim_clash_potential='this_sim_clash_potential',
            convergence_e='this_convergence_E',
            convergence_k='this_convergence_K',
            convergence_f='this_convergence_F',
            convergence_t='this_convergence_T',
            optimized_distances='this_optimized_distances')
        f_equal = ihm.flr.FPSGlobalParameters(
            forster_radius='this_forster_radius',
            conversion_function_polynom_order='this_cfpo',
            repetition='this_repetition',
            av_grid_rel='this_AV_grid_rel',
            av_min_grid_a='this_AV_min_grid_A',
            av_allowed_sphere='this_AV_allowed_sphere',
            av_search_nodes='this_AV_search_nodes',
            av_e_samples_k='this_AV_E_samples_k',
            sim_viscosity_adjustment='this_sim_viscosity_adjustment',
            sim_dt_adjustment='this_sim_dt_adjustment',
            sim_max_iter_k='this_sim_max_iter_k',
            sim_max_force='this_sim_max_force',
            sim_clash_tolerance_a='this_sim_clash_tolerance_A',
            sim_reciprocal_kt='this_sim_reciprocal_kT',
            sim_clash_potential='this_sim_clash_potential',
            convergence_e='this_convergence_E',
            convergence_k='this_convergence_K',
            convergence_f='this_convergence_F',
            convergence_t='this_convergence_T',
            optimized_distances='this_optimized_distances')
        f_unequal = ihm.flr.FPSGlobalParameters(
            forster_radius='foo',
            conversion_function_polynom_order='this_cfpo',
            repetition='this_repetition', av_grid_rel='this_AV_grid_rel',
            av_min_grid_a='this_AV_min_grid_A',
            av_allowed_sphere='this_AV_allowed_sphere',
            av_search_nodes='this_AV_search_nodes',
            av_e_samples_k='this_AV_E_samples_k',
            sim_viscosity_adjustment='this_sim_viscosity_adjustment',
            sim_dt_adjustment='this_sim_dt_adjustment',
            sim_max_iter_k='this_sim_max_iter_k',
            sim_max_force='this_sim_max_force',
            sim_clash_tolerance_a='this_sim_clash_tolerance_A',
            sim_reciprocal_kt='this_sim_reciprocal_kT',
            sim_clash_potential='this_sim_clash_potential',
            convergence_e='this_convergence_E',
            convergence_k='this_convergence_K',
            convergence_f='this_convergence_F',
            convergence_t='this_convergence_T',
            optimized_distances='this_optimized_distances')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_av_modeling_init(self):
        """ Test initialization of FPSAVModeling. """
        f = ihm.flr.FPSAVModeling(fps_modeling='foo',
                                  sample_probe='bar',
                                  parameter='foobar')
        self.assertEqual(f.fps_modeling, 'foo')
        self.assertEqual(f.sample_probe, 'bar')
        self.assertEqual(f.parameter, 'foobar')

    def test_fps_av_modeling_eq(self):
        """Test equality and inequality of FPSAVModeling objects."""
        f_ref = ihm.flr.FPSAVModeling(fps_modeling='foo',
                                      sample_probe='bar',
                                      parameter='foobar')
        f_equal = ihm.flr.FPSAVModeling(fps_modeling='foo',
                                        sample_probe='bar',
                                        parameter='foobar')
        f_unequal = ihm.flr.FPSAVModeling(fps_modeling='foo',
                                          sample_probe='bar2',
                                          parameter='foobar')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_av_parameter_init(self):
        """Test initialization of FPSAVParameter."""
        # Initialization with AV1
        f1 = ihm.flr.FPSAVParameter(num_linker_atoms='this_num_linker_atoms_1',
                                    linker_length='this_linker_length_1',
                                    linker_width='this_linker_width_1',
                                    probe_radius_1='this_probe_radius_1_1')
        self.assertEqual(f1.num_linker_atoms, 'this_num_linker_atoms_1')
        self.assertEqual(f1.linker_length, 'this_linker_length_1')
        self.assertEqual(f1.linker_width, 'this_linker_width_1')
        self.assertEqual(f1.probe_radius_1, 'this_probe_radius_1_1')
        self.assertIsNone(f1.probe_radius_2)
        self.assertIsNone(f1.probe_radius_3)
        # Initialization with AV3
        f2 = ihm.flr.FPSAVParameter(num_linker_atoms='this_num_linker_atoms_2',
                                    linker_length='this_linker_length_2',
                                    linker_width='this_linker_width_2',
                                    probe_radius_1='this_probe_radius_1_2',
                                    probe_radius_2='this_probe_radius_2_2',
                                    probe_radius_3='this_probe_radius_3_2')
        self.assertEqual(f2.num_linker_atoms, 'this_num_linker_atoms_2')
        self.assertEqual(f2.linker_length, 'this_linker_length_2')
        self.assertEqual(f2.linker_width, 'this_linker_width_2')
        self.assertEqual(f2.probe_radius_1, 'this_probe_radius_1_2')
        self.assertEqual(f2.probe_radius_2, 'this_probe_radius_2_2')
        self.assertEqual(f2.probe_radius_3, 'this_probe_radius_3_2')

    def test_fps_av_parameter_eq(self):
        """ Test equality and inequality of FPSAVParameter objects. """
        f_ref = ihm.flr.FPSAVParameter(
            num_linker_atoms='this_num_linker_atoms_1',
            linker_length='this_linker_length_1',
            linker_width='this_linker_width_1',
            probe_radius_1='this_probe_radius_1_1')
        f_equal = ihm.flr.FPSAVParameter(
            num_linker_atoms='this_num_linker_atoms_1',
            linker_length='this_linker_length_1',
            linker_width='this_linker_width_1',
            probe_radius_1='this_probe_radius_1_1')
        f_unequal = ihm.flr.FPSAVParameter(
            num_linker_atoms='this_num_linker_atoms_1', linker_length='foo',
            linker_width='this_linker_width_1',
            probe_radius_1='this_probe_radius_1_1')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_mpp_modeling_init(self):
        """Test initialization of FPSMPPModeling."""
        f = ihm.flr.FPSMPPModeling(fps_modeling='foo', mpp='bar',
                                   mpp_atom_position_group='foobar')
        self.assertEqual(f.fps_modeling, 'foo')
        self.assertEqual(f.mpp, 'bar')
        self.assertEqual(f.mpp_atom_position_group, 'foobar')

    def test_fps_mpp_modeling_eq(self):
        """Test equality and inequality of FPSMPPModeling objects."""
        f_ref = ihm.flr.FPSMPPModeling(fps_modeling='foo', mpp='bar',
                                       mpp_atom_position_group='foobar')
        f_equal = ihm.flr.FPSMPPModeling(fps_modeling='foo', mpp='bar',
                                         mpp_atom_position_group='foobar')
        f_unequal = ihm.flr.FPSMPPModeling(fps_modeling='foo2', mpp='bar',
                                           mpp_atom_position_group='foobar')

        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_mean_probe_position_init(self):
        """Test initialization of FPSMeanProbePosition."""
        f = ihm.flr.FPSMeanProbePosition(sample_probe='foo', x='bar',
                                         y='bar2', z='bar3')
        self.assertEqual(f.sample_probe, 'foo')
        self.assertEqual(f.x, 'bar')
        self.assertEqual(f.y, 'bar2')
        self.assertEqual(f.z, 'bar3')

    def test_fps_mean_probe_position_eq(self):
        """Test equality and inequality of FPSMeanProbePosition objects."""
        f_ref = ihm.flr.FPSMeanProbePosition(sample_probe='foo', x='bar',
                                             y='bar2', z='bar3')
        f_equal = ihm.flr.FPSMeanProbePosition(sample_probe='foo', x='bar',
                                               y='bar2', z='bar3')
        f_unequal = ihm.flr.FPSMeanProbePosition(
            sample_probe='foobar', x='bar', y='bar2', z='bar3')
        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_mpp_atom_position_group_init(self):
        """ Test initialization of FPSMPPAtomPositionGroup. """
        f = ihm.flr.FPSMPPAtomPositionGroup()
        self.assertEqual(f.mpp_atom_position_list, [])

    def test_fps_mpp_atom_position_group_add_atom_position(self):
        """ Test addition of an atom position to FPSMPPAtomPositionGroup. """
        f = ihm.flr.FPSMPPAtomPositionGroup()
        f.add_atom_position('foo')
        f.add_atom_position('bar')
        self.assertEqual(f.mpp_atom_position_list, ['foo', 'bar'])

    def test_fps_mpp_atom_position_group_eq(self):
        """Test equality and inequality of FPSMPPAtomPositionGroup objects."""
        f_ref = ihm.flr.FPSMPPAtomPositionGroup()
        f_ref.add_atom_position('foo')
        f_equal = ihm.flr.FPSMPPAtomPositionGroup()
        f_equal.add_atom_position('foo')
        f_unequal = ihm.flr.FPSMPPAtomPositionGroup()
        f_unequal.add_atom_position('bar')

        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_fps_mpp_atom_position_init(self):
        """Test initialization of FPSMPPAtomPosition."""
        f = ihm.flr.FPSMPPAtomPosition(atom='this_atom_id', x='this_xcoord',
                                       y='this_ycoord', z='this_zcoord')
        self.assertEqual(f.atom, 'this_atom_id')
        self.assertEqual(f.x, 'this_xcoord')
        self.assertEqual(f.y, 'this_ycoord')
        self.assertEqual(f.z, 'this_zcoord')

    def test_fps_mpp_atom_position_eq(self):
        """Test equality and inequality of FPSMPPAtomPosition objects."""
        f_ref = ihm.flr.FPSMPPAtomPosition(atom='this_atom_id',
                                           x='this_xcoord', y='this_ycoord',
                                           z='this_zcoord')
        f_equal = ihm.flr.FPSMPPAtomPosition(atom='this_atom_id',
                                             x='this_xcoord', y='this_ycoord',
                                             z='this_zcoord')
        f_unequal = ihm.flr.FPSMPPAtomPosition(
            atom='other_atom_id', x='this_xcoord', y='this_ycoord',
            z='this_zcoord')

        self.assertTrue(f_ref == f_equal)
        self.assertFalse(f_ref == f_unequal)
        self.assertTrue(f_ref != f_unequal)

    def test_flr_data_init(self):
        """ Test initialization of FLRData. """
        f = ihm.flr.FLRData()
        self.assertEqual(f.distance_restraint_groups, [])
        self.assertEqual(f.poly_probe_conjugates, [])
        self.assertEqual(f.fret_model_qualities, [])
        self.assertEqual(f.fret_model_distances, [])
        self.assertEqual(f.fps_modeling, [])

    def test_flr_data_add_distance_restraint_group(self):
        """Test addition of a distance restraint group."""
        f = ihm.flr.FLRData()
        f.distance_restraint_groups.append('foo')
        f.distance_restraint_groups.append('bar')
        self.assertEqual(f.distance_restraint_groups, ['foo', 'bar'])

    def test_flr_data_add_poly_probe_conjugate(self):
        """Test addition of a poly_probe_conjugate."""
        f = ihm.flr.FLRData()
        f.poly_probe_conjugates.extend(('foo', 'bar'))
        self.assertEqual(f.poly_probe_conjugates, ['foo', 'bar'])

    def test_flr_data_add_fret_model_quality(self):
        """Test addition of a fret_model_quality."""
        f = ihm.flr.FLRData()
        f.fret_model_qualities.extend(('foo', 'bar'))
        self.assertEqual(f.fret_model_qualities, ['foo', 'bar'])

    def test_flr_data_add_fret_model_distance(self):
        """Test addition of a fret_model_distance."""
        f = ihm.flr.FLRData()
        f.fret_model_distances.append('foo')
        f.fret_model_distances.append('bar')
        self.assertEqual(f.fret_model_distances, ['foo', 'bar'])

    def test_flr_data_add_flr_fps_modeling(self):
        """Test addition of flr_FPS_modeling."""
        f = ihm.flr.FLRData()
        f.fps_modeling.append('foo')
        f.fps_modeling.append('bar')
        self.assertEqual(f.fps_modeling, ['foo', 'bar'])

    def test_flr_data_all_chemical_descriptors(self):
        """Test for collection of all chemical descriptors."""
        f = ihm.flr.FLRData()
        # Define probe descriptors
        this_probe_descriptor_1 = ihm.flr.ProbeDescriptor(
            reactive_probe_chem_descriptor='This_reactive_probe_desc_1',
            chromophore_chem_descriptor='This_chromophore_desc_1')
        this_probe_descriptor_2 = ihm.flr.ProbeDescriptor(
            reactive_probe_chem_descriptor='This_reactive_probe_desc_2',
            chromophore_chem_descriptor='This_chromophore_desc_2')
        this_probe_descriptor_3 = ihm.flr.ProbeDescriptor(
            reactive_probe_chem_descriptor='This_reactive_probe_desc_3',
            chromophore_chem_descriptor='This_chromophore_desc_3')
        this_probe_descriptor_4 = ihm.flr.ProbeDescriptor(
            reactive_probe_chem_descriptor='Ref_reactive_probe_desc',
            chromophore_chem_descriptor='Ref_chromophore_desc')

        # Define probes
        this_probe_1 = ihm.flr.Probe(probe_list_entry='foo',
                                     probe_descriptor=this_probe_descriptor_1)
        this_probe_2 = ihm.flr.Probe(probe_list_entry='foo',
                                     probe_descriptor=this_probe_descriptor_2)
        this_probe_3 = ihm.flr.Probe(probe_list_entry='foo',
                                     probe_descriptor=this_probe_descriptor_3)
        this_probe_4 = ihm.flr.Probe(probe_list_entry='foo',
                                     probe_descriptor=this_probe_descriptor_4)

        # Define poly probe positions
        this_poly_probe_position_1 = ihm.flr.PolyProbePosition(
            resatom='foo', mutation_flag=True,
            mutated_chem_comp_id='Mutated_Chem_comp_id_1',
            modification_flag=False)
        this_poly_probe_position_2 = ihm.flr.PolyProbePosition(
            resatom='foo', mutation_flag=False, modification_flag=True,
            modified_chem_descriptor='Modified_Chem_descriptor_1')
        this_poly_probe_position_3 = ihm.flr.PolyProbePosition(
            resatom='foo', mutation_flag=False, modification_flag=False)
        this_poly_probe_position_4 = ihm.flr.PolyProbePosition(
            resatom='foo', mutation_flag=True,
            mutated_chem_comp_id='Mutated_Chem_comp_id_2')
        this_poly_probe_position_5 = ihm.flr.PolyProbePosition(
            resatom='foo', modification_flag=True,
            modified_chem_descriptor='Modified_Chem_descriptor_2')

        this_sample_probe_1 = ihm.flr.SampleProbeDetails(
            sample='foo', probe=this_probe_1, fluorophore_type='donor',
            poly_probe_position=this_poly_probe_position_1)
        this_sample_probe_2 = ihm.flr.SampleProbeDetails(
            sample='foo2', probe=this_probe_2, fluorophore_type='donor',
            poly_probe_position=this_poly_probe_position_2)
        this_sample_probe_3 = ihm.flr.SampleProbeDetails(
            sample='foo3', probe=this_probe_3, fluorophore_type='donor',
            poly_probe_position=this_poly_probe_position_3)
        this_sample_probe_4 = ihm.flr.SampleProbeDetails(
            sample='foo4', probe=this_probe_4, fluorophore_type='donor',
            poly_probe_position=this_poly_probe_position_4)
        this_sample_probe_5 = ihm.flr.SampleProbeDetails(
            sample='foo5', probe=this_probe_4, fluorophore_type='donor',
            poly_probe_position=this_poly_probe_position_5)

        this_reference_measurement_1 = ihm.flr.RefMeasurement(
            ref_sample_probe=this_sample_probe_4)
        this_reference_measurement_2 = ihm.flr.RefMeasurement(
            ref_sample_probe=this_sample_probe_5)
        this_reference_measurement_group_1 = ihm.flr.RefMeasurementGroup()
        this_reference_measurement_group_1.add_ref_measurement(
            this_reference_measurement_1)
        this_reference_measurement_group_1.add_ref_measurement(
            this_reference_measurement_2)

        this_analysis_1 = ihm.flr.FRETAnalysis(
            experiment='foo', sample_probe_1=this_sample_probe_1,
            sample_probe_2=this_sample_probe_2, forster_radius='foo',
            type='lifetime-based',
            ref_measurement_group=this_reference_measurement_group_1)
        this_analysis_2 = ihm.flr.FRETAnalysis(
            experiment='bar', sample_probe_1=this_sample_probe_1,
            sample_probe_2=this_sample_probe_2, forster_radius='bar',
            type='intensity-based')

        this_distance_restraint_1 = ihm.flr.FRETDistanceRestraint(
            sample_probe_1=this_sample_probe_1,
            sample_probe_2=this_sample_probe_2,
            analysis=this_analysis_1, distance=50)
        this_distance_restraint_2 = ihm.flr.FRETDistanceRestraint(
            sample_probe_1=this_sample_probe_1,
            sample_probe_2=this_sample_probe_3,
            analysis=this_analysis_2, distance=50)

        this_distance_restraint_group = ihm.flr.FRETDistanceRestraintGroup()
        this_distance_restraint_group.add_distance_restraint(
            this_distance_restraint_1)
        this_distance_restraint_group.add_distance_restraint(
            this_distance_restraint_2)
        f.distance_restraint_groups.append(this_distance_restraint_group)

        conj = ihm.flr.PolyProbeConjugate(
            sample_probe=this_sample_probe_1,
            chem_descriptor='Conjugate_probe_desc',
            ambiguous_stoichiometry=False)
        f.poly_probe_conjugates.append(conj)

        descs = list(f._all_flr_chemical_descriptors())
        self.assertEqual(
            descs,
            ['This_reactive_probe_desc_1', 'This_chromophore_desc_1',
             'This_reactive_probe_desc_2', 'This_chromophore_desc_2',
             'Modified_Chem_descriptor_1',
             'Ref_reactive_probe_desc', 'Ref_chromophore_desc',
             'Ref_reactive_probe_desc', 'Ref_chromophore_desc',
             'Modified_Chem_descriptor_2',
             'This_reactive_probe_desc_1', 'This_chromophore_desc_1',
             'This_reactive_probe_desc_3', 'This_chromophore_desc_3',
             'Conjugate_probe_desc'])


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.citations


class Tests(unittest.TestCase):

    def test_citations(self):
        """Test citations module"""
        pmi = ihm.citations.pmi
        self.assertEqual(pmi.pmid, '31396911')
        self.assertEqual(pmi.doi, '10.1007/978-1-4939-9608-7_15')


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.analysis


class Tests(unittest.TestCase):

    def test_filter_step(self):
        """Test analysis FilterStep class"""
        s = ihm.analysis.FilterStep(feature='RMSD', num_models_begin=42,
                                    num_models_end=5)
        self.assertEqual(s.type, 'filter')
        self.assertEqual(s.feature, 'RMSD')
        self.assertEqual(s.num_models_begin, 42)
        self.assertEqual(s.num_models_end, 5)

    def test_cluster_step(self):
        """Test analysis ClusterStep class"""
        s = ihm.analysis.ClusterStep(feature='RMSD', num_models_begin=42,
                                     num_models_end=5)
        self.assertEqual(s.type, 'cluster')
        self.assertEqual(s.feature, 'RMSD')
        self.assertEqual(s.num_models_begin, 42)
        self.assertEqual(s.num_models_end, 5)

    def test_rescore_step(self):
        """Test analysis RescoreStep class"""
        s = ihm.analysis.RescoreStep(feature='RMSD', num_models_begin=42,
                                     num_models_end=5)
        self.assertEqual(s.type, 'rescore')
        self.assertEqual(s.feature, 'RMSD')
        self.assertEqual(s.num_models_begin, 42)
        self.assertEqual(s.num_models_end, 5)

    def test_step(self):
        """Test analysis Step class"""
        s = ihm.analysis.Step(feature='RMSD', num_models_begin=42,
                              num_models_end=5)
        self.assertEqual(s.type, 'other')
        self.assertEqual(s.feature, 'RMSD')
        self.assertEqual(s.num_models_begin, 42)
        self.assertEqual(s.num_models_end, 5)

    def test_empty_step(self):
        """Test analysis EmptyStep class"""
        s = ihm.analysis.EmptyStep()
        self.assertEqual(s.type, 'none')
        self.assertEqual(s.feature, 'none')
        self.assertIsNone(s.num_models_begin)
        self.assertIsNone(s.num_models_end)

    def test_analysis(self):
        """Test Analysis class"""
        a = ihm.analysis.Analysis()
        self.assertEqual(a.steps, [])


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.restraint


class Tests(unittest.TestCase):

    def test_restraint(self):
        """Test Restraint base class"""
        _ = ihm.restraint.Restraint()  # does nothing

    def test_em3d_restraint_fit(self):
        """Test EM3DRestraintFit class"""
        f = ihm.restraint.EM3DRestraintFit(0.4)
        self.assertAlmostEqual(f.cross_correlation_coefficient, 0.4, delta=0.1)

    def test_em3d_restraint(self):
        """Test EM3DRestraint class"""
        f = ihm.restraint.EM3DRestraint(dataset='foo', assembly='bar')
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.assembly, 'bar')
        self.assertEqual(f.fits, {})

    def test_sas_restraint_fit(self):
        """Test SASRestraintFit class"""
        f = ihm.restraint.SASRestraintFit(0.4)
        self.assertAlmostEqual(f.chi_value, 0.4, delta=0.1)

    def test_sas_restraint(self):
        """Test SASRestraint class"""
        f = ihm.restraint.SASRestraint(dataset='foo', assembly='bar')
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.assembly, 'bar')
        self.assertEqual(f.fits, {})

    def test_em2d_restraint_fit(self):
        """Test EM2DRestraintFit class"""
        f = ihm.restraint.EM2DRestraintFit(0.4)
        self.assertAlmostEqual(f.cross_correlation_coefficient, 0.4, delta=0.1)
        self.assertIsNone(f.rot_matrix)
        self.assertIsNone(f.tr_vector)

    def test_em2d_restraint(self):
        """Test EM2DRestraint class"""
        f = ihm.restraint.EM2DRestraint(dataset='foo', assembly='bar')
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.assembly, 'bar')
        self.assertEqual(f.fits, {})

    def test_distance_restraint(self):
        """Test DistanceRestraint class"""
        _ = ihm.restraint.DistanceRestraint()  # does nothing

    def test_harmonic_distance_restraint(self):
        """Test HarmonicDistanceRestraint class"""
        r = ihm.restraint.HarmonicDistanceRestraint(42.0)
        self.assertAlmostEqual(r.distance, 42.0, delta=0.1)
        self.assertAlmostEqual(r.distance_lower_limit, 42.0, delta=0.1)
        self.assertAlmostEqual(r.distance_upper_limit, 42.0, delta=0.1)
        self.assertEqual(r.restraint_type, "harmonic")

    def test_upper_bound_distance_restraint(self):
        """Test UpperBoundDistanceRestraint class"""
        r = ihm.restraint.UpperBoundDistanceRestraint(42.0)
        self.assertAlmostEqual(r.distance, 42.0, delta=0.1)
        self.assertIsNone(r.distance_lower_limit)
        self.assertAlmostEqual(r.distance_upper_limit, 42.0, delta=0.1)
        self.assertEqual(r.restraint_type, "upper bound")

    def test_lower_bound_distance_restraint(self):
        """Test LowerBoundDistanceRestraint class"""
        r = ihm.restraint.LowerBoundDistanceRestraint(42.0)
        self.assertAlmostEqual(r.distance, 42.0, delta=0.1)
        self.assertAlmostEqual(r.distance_lower_limit, 42.0, delta=0.1)
        self.assertIsNone(r.distance_upper_limit)
        self.assertEqual(r.restraint_type, "lower bound")

    def test_lower_upper_bound_distance_restraint(self):
        """Test LowerUpperBoundDistanceRestraint class"""
        r = ihm.restraint.LowerUpperBoundDistanceRestraint(20.0, 30.0)
        self.assertAlmostEqual(r.distance_lower_limit, 20.0, delta=0.1)
        self.assertAlmostEqual(r.distance_upper_limit, 30.0, delta=0.1)
        self.assertEqual(r.restraint_type, "lower and upper bound")

    def test_cross_link_restraint(self):
        """Test CrossLinkRestraint class"""
        dss = ihm.ChemDescriptor('DSS')
        f = ihm.restraint.CrossLinkRestraint(dataset='foo', linker=dss)
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.linker, dss)
        self.assertEqual(f.experimental_cross_links, [])

    def test_experimental_cross_link(self):
        """Test ExperimentalCrossLink class"""
        f = ihm.restraint.ExperimentalCrossLink('res1', 'res2')
        self.assertEqual(f.residue1, 'res1')
        self.assertEqual(f.residue2, 'res2')

    def test_cross_link(self):
        """Test CrossLink class"""
        _ = ihm.restraint.CrossLink()  # does nothing

    def test_residue_cross_link(self):
        """Test ResidueCrossLink class"""
        f = ihm.restraint.ResidueCrossLink(
            experimental_cross_link='ex', asym1='asym1', asym2='asym2',
            distance='dist')
        self.assertEqual(f.granularity, 'by-residue')
        self.assertIsNone(f.atom1)
        self.assertIsNone(f.atom2)
        self.assertEqual(f.asym1, 'asym1')
        self.assertEqual(f.asym2, 'asym2')

    def test_feature_cross_link(self):
        """Test FeatureCrossLink class"""
        f = ihm.restraint.FeatureCrossLink(
            experimental_cross_link='ex', asym1='asym1', asym2='asym2',
            distance='dist')
        self.assertEqual(f.granularity, 'by-feature')
        self.assertIsNone(f.atom1)
        self.assertIsNone(f.atom2)
        self.assertEqual(f.asym1, 'asym1')
        self.assertEqual(f.asym2, 'asym2')

    def test_atom_cross_link(self):
        """Test AtomCrossLink class"""
        f = ihm.restraint.AtomCrossLink(
            experimental_cross_link='ex', asym1='asym1', asym2='asym2',
            atom1='C', atom2='N', distance='dist')
        self.assertEqual(f.granularity, 'by-atom')
        self.assertEqual(f.atom1, 'C')
        self.assertEqual(f.atom2, 'N')
        self.assertEqual(f.asym1, 'asym1')
        self.assertEqual(f.asym2, 'asym2')

    def test_feature(self):
        """Test Feature base class"""
        f = ihm.restraint.Feature()  # does nothing
        self.assertEqual(f._all_entities_or_asyms(), [])

    def test_residue_feature(self):
        """Test ResidueFeature class"""
        e = ihm.Entity('AHCDAH')
        a = ihm.AsymUnit(e)
        f = ihm.restraint.ResidueFeature(ranges=[])
        self.assertIsNone(f._get_entity_type())

        # No ranges - type is 'residue'
        f = ihm.restraint.ResidueFeature(ranges=[])
        self.assertEqual(f.type, 'residue')

        # All ranges are individual residues
        f = ihm.restraint.ResidueFeature(ranges=[a(1, 1), a(2, 2)])
        self.assertEqual(f.type, 'residue')

        # Should work with actual Residue objects too
        f = ihm.restraint.ResidueFeature(ranges=[a.residue(3)])
        self.assertEqual(f.type, 'residue')

        # At least one range is a true range
        f = ihm.restraint.ResidueFeature(ranges=[a(3, 4)])
        self.assertEqual(f.type, 'residue range')

    def test_geometric_restraint(self):
        """Test GeometricRestraint class"""
        f = ihm.restraint.GeometricRestraint(
            dataset='foo', geometric_object='geom', feature='feat',
            distance='dist')
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.object_characteristic, 'other')
        self.assertIsNone(f.assembly)

    def test_center_geometric_restraint(self):
        """Test CenterGeometricRestraint class"""
        f = ihm.restraint.CenterGeometricRestraint(
            dataset='foo', geometric_object='geom', feature='feat',
            distance='dist')
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.object_characteristic, 'center')
        self.assertIsNone(f.assembly)

    def test_inner_surface_geometric_restraint(self):
        """Test InnerSurfaceGeometricRestraint class"""
        f = ihm.restraint.InnerSurfaceGeometricRestraint(
            dataset='foo', geometric_object='geom', feature='feat',
            distance='dist')
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.object_characteristic, 'inner surface')
        self.assertIsNone(f.assembly)

    def test_outer_surface_geometric_restraint(self):
        """Test OuterSurfaceGeometricRestraint class"""
        f = ihm.restraint.OuterSurfaceGeometricRestraint(
            dataset='foo', geometric_object='geom', feature='feat',
            distance='dist')
        self.assertEqual(f.dataset, 'foo')
        self.assertEqual(f.object_characteristic, 'outer surface')
        self.assertIsNone(f.assembly)
        self.assertEqual(f.feature, 'feat')
        self.assertEqual(f._all_features, ('feat',))

    def test_derived_distance_restraint(self):
        """Test DerivedDistanceRestraint class"""
        f = ihm.restraint.DerivedDistanceRestraint(
            dataset='foo', feature1='feat1', feature2='feat2', distance='dist')
        self.assertEqual(f.dataset, 'foo')
        self.assertIsNone(f.assembly)
        self.assertEqual(f.feature1, 'feat1')
        self.assertEqual(f.feature2, 'feat2')
        self.assertEqual(f._all_features, ('feat1', 'feat2'))


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest
import sys
import subprocess
import pickle

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)

import ihm.reader


def get_example_dir():
    return os.path.join(TOPDIR, "examples")


def get_example_path(fname):
    return os.path.join(get_example_dir(), fname)


class Tests(unittest.TestCase):

    @unittest.skipIf('APPVEYOR' in os.environ,
                     "AppVeyor environments have old SSL certs")
    @unittest.skipIf('GITHUB_ACTIONS' in os.environ,
                     "Example is slow and fails when PDB-Dev is down")
    def test_validator_example(self):
        """Test validator example"""
        subprocess.check_call([sys.executable,
                               get_example_path("validate_pdb_dev.py")])

    def test_simple_docking_example(self):
        """Test simple-docking example"""
        with utils.temporary_directory() as tmpdir:
            subprocess.check_call([sys.executable,
                                   get_example_path("simple-docking.py")],
                                  cwd=tmpdir)

            # Make sure that a complete output file was produced and that we
            # can read it
            with open(os.path.join(tmpdir, 'output.cif')) as fh:
                contents = fh.readlines()
            self.assertEqual(len(contents), 315)
            with open(os.path.join(tmpdir, 'output.cif')) as fh:
                s, = ihm.reader.read(fh)

    def test_locations_example(self):
        """Test locations example"""
        subprocess.check_call([sys.executable, "locations.py"],
                              cwd=get_example_dir())
        out = get_example_path("output.cif")

        # Make sure that a complete output file was produced and that we
        # can read it
        with open(out) as fh:
            contents = fh.readlines()
        self.assertEqual(len(contents), 69)
        with open(out) as fh:
            s, = ihm.reader.read(fh)
        os.unlink(out)

    def test_ligands_water_example(self):
        """Test ligands_water example"""
        subprocess.check_call([sys.executable, "ligands_water.py"],
                              cwd=get_example_dir())
        out = get_example_path("output.cif")

        # Make sure that a complete output file was produced and that we
        # can read it
        with open(out) as fh:
            contents = fh.readlines()
        self.assertEqual(len(contents), 161)
        with open(out) as fh:
            s, = ihm.reader.read(fh)
        # Make sure that resulting Python objects are picklable
        testpck = 'test-lig-wat.pck'
        with open(testpck, 'wb') as fh:
            pickle.dump(s, fh, protocol=-1)
        with open(testpck, 'rb') as fh:
            _ = pickle.load(fh)
        os.unlink(out)
        os.unlink(testpck)

    def test_non_standard_residues_example(self):
        """Test non_standard_residues example"""
        subprocess.check_call([sys.executable, "non_standard_residues.py"],
                              cwd=get_example_dir())
        out = get_example_path("output.cif")

        # Make sure that a complete output file was produced and that we
        # can read it
        with open(out) as fh:
            contents = fh.readlines()
        self.assertEqual(len(contents), 65)
        with open(out) as fh:
            s, = ihm.reader.read(fh)
        os.unlink(out)

    def test_stream_parser_example(self):
        """Test stream_parser example"""
        subprocess.check_call([sys.executable, "stream_parser.py"],
                              cwd=get_example_dir())


if __name__ == '__main__':
    unittest.main()



# This example demonstrates how non-polymeric entities (ligands, water)
# are handled by the Python IHM library. See the simple-docking.py example
# for an introduction to the library.

import ihm
import ihm.dumper

system = ihm.System()

# An entity corresponding to an amino acid (polyalanine) sequence
entity_protein = ihm.Entity('AAA', description='Subunit A')

# An entity corresponding to an RNA sequence
entity_rna = ihm.Entity('ACG', alphabet=ihm.RNAAlphabet,
                        description='RNA chain')

# An entity corresponding to a DNA sequence
entity_dna = ihm.Entity(['DA', 'DC', 'DG'], alphabet=ihm.DNAAlphabet,
                        description='DNA chain')

# Non-polymers such as ligands or water should each live in their own Entity:

# A ligand entity (in this case, heme)
heme = ihm.NonPolymerChemComp("HEM", name='PROTOPORPHYRIN IX CONTAINING FE',
                              formula='C34 H32 Fe N4 O4')
entity_heme = ihm.Entity([heme], description='Heme')

# Water
entity_h2o = ihm.Entity([ihm.WaterChemComp()], description='Water')

system.entities.extend((entity_protein, entity_rna, entity_dna, entity_heme,
                        entity_h2o))

# Next, we define asymmetric units for everything we modeled.
# Here, we have a single instance of each protein, RNA and DNA, two hemes,
# plus crystal waters
asym_protein = ihm.AsymUnit(entity_protein, details='Subunit A')
asym_rna = ihm.AsymUnit(entity_rna, details='RNA chain')
asym_dna = ihm.AsymUnit(entity_dna, details='DNA chain')
asym_heme1 = ihm.AsymUnit(entity_heme, details='First heme')
asym_heme2 = ihm.AsymUnit(entity_heme, details='Second heme')
asym_h2o = ihm.AsymUnit(entity_h2o, details='Crystal waters')
system.asym_units.extend((asym_protein, asym_rna, asym_dna, asym_heme1,
                          asym_heme2, asym_h2o))

# todo: show handling of multiple waters

# Once the system is complete, we can write it out to an mmCIF file:
with open('output.cif', 'w') as fh:
    ihm.dumper.write(fh, [system])



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.geometry


class Tests(unittest.TestCase):

    def test_center(self):
        """Test Center class"""
        c = ihm.geometry.Center(1, 2, 3)
        self.assertEqual(c.x, 1)
        self.assertEqual(c.y, 2)
        self.assertEqual(c.z, 3)

    def test_transformation(self):
        """Test Transformation class"""
        _ = ihm.geometry.Transformation([[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                                        [1, 2, 3])

    def test_identity_transformation(self):
        """Test identity transformation"""
        t = ihm.geometry.Transformation.identity()
        for i in range(3):
            self.assertAlmostEqual(t.tr_vector[i], 0., delta=0.1)
            for j in range(3):
                self.assertAlmostEqual(t.rot_matrix[i][j],
                                       1. if i == j else 0., delta=0.1)

    def test_geometric_object(self):
        """Test GeometricObject bass class"""
        g = ihm.geometry.GeometricObject(name='foo', description='bar')
        self.assertEqual(g.name, 'foo')
        self.assertEqual(g.description, 'bar')
        self.assertEqual(g.type, 'other')

    def test_sphere(self):
        """Test Sphere class"""
        c = ihm.geometry.Center(1, 2, 3)
        t = ihm.geometry.Transformation([[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                                        [1, 2, 3])
        g = ihm.geometry.Sphere(center=c, transformation=t, radius=5)
        self.assertIsNone(g.name)
        self.assertEqual(g.radius, 5)
        self.assertEqual(g.type, 'sphere')
        g = ihm.geometry.Sphere(center=c, radius=5)
        self.assertIsNone(g.transformation)

    def test_torus(self):
        """Test Torus class"""
        c = ihm.geometry.Center(1, 2, 3)
        t = ihm.geometry.Transformation([[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                                        [1, 2, 3])
        g = ihm.geometry.Torus(center=c, transformation=t, major_radius=5,
                               minor_radius=1)
        self.assertIsNone(g.name)
        self.assertEqual(g.major_radius, 5)
        self.assertEqual(g.type, 'torus')
        g = ihm.geometry.Torus(center=c, major_radius=5,
                               minor_radius=1)
        self.assertIsNone(g.transformation)

    def test_half_torus(self):
        """Test HalfTorus class"""
        c = ihm.geometry.Center(1, 2, 3)
        t = ihm.geometry.Transformation([[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                                        [1, 2, 3])
        g = ihm.geometry.HalfTorus(center=c, transformation=t, major_radius=5,
                                   minor_radius=1, thickness=0.1, inner=True)
        self.assertIsNone(g.name)
        self.assertEqual(g.major_radius, 5)
        self.assertEqual(g.inner, True)
        self.assertEqual(g.type, 'half-torus')
        g = ihm.geometry.HalfTorus(center=c, major_radius=5,
                                   minor_radius=1, thickness=0.1, inner=True)
        self.assertIsNone(g.transformation)

    def test_axis(self):
        """Test Axis base class"""
        g = ihm.geometry.Axis(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'axis')

    def test_x_axis(self):
        """Test XAxis class"""
        g = ihm.geometry.XAxis(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'axis')
        self.assertEqual(g.axis_type, 'x-axis')

    def test_y_axis(self):
        """Test YAxis class"""
        g = ihm.geometry.YAxis(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'axis')
        self.assertEqual(g.axis_type, 'y-axis')

    def test_z_axis(self):
        """Test ZAxis class"""
        g = ihm.geometry.ZAxis(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'axis')
        self.assertEqual(g.axis_type, 'z-axis')

    def test_plane(self):
        """Test Plane base class"""
        g = ihm.geometry.Plane(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'plane')

    def test_xy_plane(self):
        """Test XYPlane base class"""
        g = ihm.geometry.XYPlane(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'plane')
        self.assertEqual(g.plane_type, 'xy-plane')

    def test_yz_plane(self):
        """Test YZPlane base class"""
        g = ihm.geometry.YZPlane(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'plane')
        self.assertEqual(g.plane_type, 'yz-plane')

    def test_xz_plane(self):
        """Test XZPlane base class"""
        g = ihm.geometry.XZPlane(name='foo', description='bar')
        self.assertIsNone(g.transformation)
        self.assertEqual(g.type, 'plane')
        self.assertEqual(g.plane_type, 'xz-plane')


if __name__ == '__main__':
    unittest.main()



# This example demonstrates a variety of ways an IHM mmCIF file can point
# to external resources, such as script files, modeling trajectories, or
# electron microscopy density maps. These may be too big to fit efficiently
# in the mmCIF file, or may already be deposited in an experiment-specific
# database (so it makes no sense to copy them).

import ihm.location
import ihm.dataset
import ihm.dumper

system = ihm.System()

# To point to an external file, we use one of the classes in the ihm.location
# module. Here we reference this Python script itself on the local disk (output
# paths in the mmCIF file will be relative to the current working directory):
loc = ihm.location.WorkflowFileLocation(
    "locations.py",
    details="The Python script used to generate this mmCIF file")
# Add the location to the system, so it gets output to the mmCIF file
system.locations.append(loc)

# For public mmCIF files, external files need to also be in a public location,
# for example, in an archive file stored at a service such as Zenodo that
# assigns a DOI. To handle this, we use a Repository object:
r = ihm.location.Repository(
    doi='10.5281/zenodo.820724',
    url='https://zenodo.org/record/820724/files/archive.zip')
loc = ihm.location.OutputFileLocation("densities/subunitA.mrc", repo=r)
system.locations.append(loc)

# Users of the mmCIF can then obtain the file subunitA.mrc by downloading
# archive.zip from the given DOI or URL, unzipping it, and then looking in the
# densities directory. Multiple files can share the same repository.

# Note that this URL is for example purposes only (there isn't really an
# 'archive.zip' at that DOI).

# Datasets are the most common users of external files. For example, to refer
# to an input PDB file in the current directory:
loc = ihm.location.InputFileLocation("simple.pdb", details="Input PDB file")
d = ihm.dataset.PDBDataset(loc)
# Add the dataset to the mmCIF file. (Normally a dataset would be added to the
# object that uses it, such as a restraint. If we want to include a dataset
# that isn't referenced from anything else, as in this example, we can add it
# to the 'orphan' list.)
system.orphan_datasets.append(d)

# Generally, datasets will be deposited in an experiment-specific database.
# We can point to such a database using a subclass of DatabaseLocation, for
# example to point to PDB:
loc = ihm.location.PDBLocation('1abc')
system.orphan_datasets.append(ihm.dataset.PDBDataset(loc))

# If the current working directory is itself a checkout of a repository which
# is archived at a DOI, we can retroactively update all 'local' paths added
# above to point to this DOI. After calling update_locations_in_repositories(),
# all files under the parent directory (..) are assumed to be available in
# the python-ihm.zip archive. For example, simple.pdb can be found as
# python-ihm-v0.1/examples/simple.pdb in the archive.
r = ihm.location.Repository(
    doi='10.5281/zenodo.802915',
    url='https://zenodo.org/record/802915/files/python-ihm.zip',
    top_directory="python-ihm-v0.1", root="..")
system.update_locations_in_repositories([r])

# Write out everything to an mmCIF file
with open('output.cif', 'w') as fh:
    ihm.dumper.write(fh, [system])



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.util


class Tests(unittest.TestCase):
    def test_asym_ids(self):
        """Test _AsymIDs class"""
        c = ihm.util._AsymIDs()
        self.assertEqual([c[i] for i in range(0, 4)],
                         ['A', 'B', 'C', 'D'])
        self.assertEqual([c[i] for i in range(24, 28)],
                         ['Y', 'Z', 'AA', 'AB'])
        self.assertEqual([c[i] for i in range(50, 54)],
                         ['AY', 'AZ', 'BA', 'BB'])
        self.assertEqual([c[i] for i in range(700, 704)],
                         ['ZY', 'ZZ', 'AAA', 'AAB'])

    def test_remove_id(self):
        """Test _remove_id utility function"""
        class DummyObj(object):
            pass
        o = DummyObj()
        ihm.util._remove_id(o)
        o._id = 42
        ihm.util._remove_id(o)
        self.assertFalse(hasattr(o, '_id'))

    def test_assign_id(self):
        """Test _assign_id utility function"""
        class DummyObj(object):
            def __init__(self, hashval):
                self.hashval = hashval

            def __eq__(self, other):
                return self.hashval == other.hashval

            def __hash__(self):
                return self.hashval
        seen_objs = {}
        obj_by_id = []
        obj1a = DummyObj(42)  # obj1a and 1b are identical
        obj1b = DummyObj(42)
        obj2 = DummyObj(34)
        obj3 = DummyObj(23)  # obj3 already has an id
        obj3._id = 'foo'
        for obj in (obj1a, obj1b, obj2, obj3):
            ihm.util._assign_id(obj, seen_objs, obj_by_id)
        self.assertEqual(obj1a._id, 1)
        self.assertEqual(obj1b._id, 1)
        self.assertEqual(obj2._id, 2)
        self.assertEqual(obj3._id, 'foo')
        self.assertEqual(obj_by_id, [obj1a, obj2])

    def test_get_relative_path(self):
        """Test get_relative_path()"""
        if os.path.sep == '/':
            self.assertEqual(ihm.util._get_relative_path('/foo/bar',
                                                         '/foo/bar'),
                             '/foo/bar')
            self.assertEqual(ihm.util._get_relative_path('/foo/bar', 'baz'),
                             '/foo/baz')


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.representation


class Tests(unittest.TestCase):

    def test_segment(self):
        """Test Segment base class"""
        _ = ihm.representation.Segment()  # does nothing

    def test_atomic_segment(self):
        """Test AtomicSegment class"""
        asym = ihm.AsymUnit(ihm.Entity('A' * 30))
        s = ihm.representation.AtomicSegment(
            asym_unit=asym(1, 10), rigid=True, starting_model=None)
        self.assertEqual(s.asym_unit.seq_id_range, (1, 10))
        self.assertEqual(s.primitive, 'atomistic')
        self.assertEqual(s.granularity, 'by-atom')
        self.assertIsNone(s.count)
        self.assertEqual(s.rigid, True)

    def test_residue_segment(self):
        """Test ResidueSegment class"""
        asym = ihm.AsymUnit(ihm.Entity('AAAA'))
        s = ihm.representation.ResidueSegment(
            asym_unit=asym, rigid=True, primitive='sphere')
        self.assertEqual(s.asym_unit.seq_id_range, (1, 4))
        self.assertEqual(s.primitive, 'sphere')
        self.assertEqual(s.granularity, 'by-residue')
        self.assertIsNone(s.count)
        self.assertEqual(s.rigid, True)

    def test_multi_residue_segment(self):
        """Test MultiResidueSegment class"""
        asym = ihm.AsymUnit(ihm.Entity('AAAA'))
        s = ihm.representation.MultiResidueSegment(
            asym_unit=asym, rigid=True, primitive='sphere')
        self.assertEqual(s.primitive, 'sphere')
        self.assertEqual(s.granularity, 'multi-residue')
        self.assertIsNone(s.count)
        self.assertEqual(s.rigid, True)

    def test_feature_segment(self):
        """Test FeatureSegment class"""
        asym = ihm.AsymUnit(ihm.Entity('AAAA'))
        s = ihm.representation.FeatureSegment(
            asym_unit=asym, rigid=True, primitive='sphere', count=4)
        self.assertEqual(s.primitive, 'sphere')
        self.assertEqual(s.granularity, 'by-feature')
        self.assertEqual(s.count, 4)
        self.assertEqual(s.rigid, True)

    def test_representation(self):
        """Test Representation class"""
        asym = ihm.AsymUnit(ihm.Entity('AAAA'))
        s = ihm.representation.AtomicSegment(asym_unit=asym, rigid=True)
        r = ihm.representation.Representation()
        r.append(s)
        self.assertEqual(len(r), 1)


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest
import sys

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.format_bcif


# Provide dummy implementations of msgpack.unpack() and msgpack.pack() which
# just return the data unchanged. We can use these to test the BinaryCIF
# parser with Python objects rather than having to install msgpack and
# generate real binary files
class MockMsgPack(object):
    @staticmethod
    def unpack(fh, raw=False):
        return fh

    @staticmethod
    def pack(data, fh, use_bin_type=True):
        fh.data = data


class MockFh(object):
    pass


class GenericHandler(object):
    """Capture BinaryCIF data as a simple list of dicts"""
    not_in_file = None
    omitted = None
    unknown = "?"

    _keys = ('method', 'foo', 'bar', 'baz', 'pdbx_keywords', 'var1',
             'var2', 'var3')

    def __init__(self):
        self.data = []

    def __call__(self, *args):
        d = {}
        for k, v in zip(self._keys, args):
            if v is not None:
                d[k] = v
        self.data.append(d)


def _encode(rows):
    # Assume rows is a list of strings; make simple BinaryCIF encoding
    mask = [0] * len(rows)
    need_mask = False
    for i, row in enumerate(rows):
        if row is None:
            need_mask = True
            mask[i] = 1
        elif row == '?':
            need_mask = True
            mask[i] = 2
    if need_mask:
        rows = ['' if r == '?' or r is None else r for r in rows]
        mask = {u'data': ''.join(chr(i) for i in mask).encode('ascii'),
                u'encoding': [{u'kind': u'ByteArray',
                               u'type': ihm.format_bcif._Uint8}]}
    else:
        mask = None
    string_data = "".join(rows)

    offsets = []
    total_len = 0
    for row in rows:
        offsets.append(total_len)
        total_len += len(row)
    offsets.append(total_len)
    offsets = ''.join(chr(i) for i in offsets).encode('ascii')
    indices = ''.join(chr(i) for i in range(len(rows))).encode('ascii')
    string_array_encoding = {
        u'kind': u'StringArray',
        u'dataEncoding': [{u'kind': u'ByteArray',
                           u'type': ihm.format_bcif._Uint8}],
        u'stringData': string_data,
        u'offsetEncoding': [{u'kind': u'ByteArray',
                             u'type': ihm.format_bcif._Uint8}],
        u'offsets': offsets}
    d = {u'data': indices,
         u'encoding': [string_array_encoding]}
    return d, mask


class Category(object):
    def __init__(self, name, data):
        self.name = name
        self.data = data

    def get_bcif(self):
        nrows = 0
        cols = []
        for name, rows in self.data.items():
            nrows = len(rows)
            data, mask = _encode(rows)
            cols.append({u'mask': mask, u'name': name,
                         u'data': data})
        return {u'name': self.name,
                u'columns': cols, u'rowCount': nrows}


class Block(list):
    pass


def _make_bcif_file(blocks):
    blocks = [{u'header': 'ihm',
               u'categories': [c.get_bcif() for c in block]}
              for block in blocks]
    return {u'version': u'0.1', u'encoder': u'python-ihm test suite',
            u'dataBlocks': blocks}


class Tests(unittest.TestCase):
    def test_decode_bytes(self):
        """Test decode_bytes function"""
        d = ihm.format_bcif._decode_bytes(u'foo')
        self.assertEqual(d, 'foo')

    def test_decoder_base(self):
        """Test Decoder base class"""
        d = ihm.format_bcif._Decoder()
        self.assertIsNone(d._kind)
        d(enc=None, data=None)  # noop

    def test_string_array_decoder(self):
        """Test StringArray decoder"""
        d = ihm.format_bcif._StringArrayDecoder()
        self.assertEqual(d._kind, u'StringArray')

        # Int8 is signed char (so FF is -1)
        enc = {u'stringData': u'aAB',
               u'dataEncoding': [{u'kind': u'ByteArray',
                                  u'type': ihm.format_bcif._Int8}],
               u'offsetEncoding': [{u'kind': u'ByteArray',
                                    u'type': ihm.format_bcif._Int8}],
               u'offsets': b'\x00\x01\x03'}
        data = b'\x00\x01\x00\xFF'

        data = d(enc, data)
        self.assertEqual(list(data), ['a', 'AB', 'a', None])

    def test_byte_array_decoder(self):
        """Test ByteArray decoder"""
        d = ihm.format_bcif._ByteArrayDecoder()
        self.assertEqual(d._kind, u'ByteArray')

        # type 1 (signed char)
        data = d({u'type': ihm.format_bcif._Int8}, b'\x00\x01\xFF')
        self.assertEqual(list(data), [0, 1, -1])

        # type 2 (signed short)
        data = d({u'type': ihm.format_bcif._Int16}, b'\x00\x01\x01\xAC')
        self.assertEqual(list(data), [256, -21503])

        # type 3 (signed int)
        data = d({u'type': ihm.format_bcif._Int32}, b'\x00\x01\x01\x05')
        self.assertEqual(list(data), [83951872])

        # type 4 (unsigned char)
        data = d({u'type': ihm.format_bcif._Uint8}, b'\x00\xFF')
        self.assertEqual(list(data), [0, 255])

        # type 5 (unsigned short)
        data = d({u'type': ihm.format_bcif._Uint16}, b'\x00\x01\x01\xAC')
        self.assertEqual(list(data), [256, 44033])

        # type 6 (unsigned int)
        data = d({u'type': ihm.format_bcif._Uint32}, b'\x00\x01\x01\xFF')
        self.assertEqual(list(data), [4278255872])

        # type 32 (32-bit float)
        data = d({u'type': ihm.format_bcif._Float32}, b'\x00\x00(B')
        self.assertAlmostEqual(list(data)[0], 42.0, delta=0.1)

        # type 33 (64-bit float)
        data = d({u'type': ihm.format_bcif._Float64},
                 b'\x00\x00\x00\x00\x00\x00E@')
        self.assertAlmostEqual(list(data)[0], 42.0, delta=0.1)

    def test_integer_packing_decoder_signed(self):
        """Test IntegerPacking decoder with signed data"""
        d = ihm.format_bcif._IntegerPackingDecoder()
        self.assertEqual(d._kind, u'IntegerPacking')

        # 1-byte data
        data = d({u'isUnsigned': False, u'byteCount': 1},
                 [1, 2, -3, 127, 1, -128, -5])
        self.assertEqual(list(data), [1, 2, -3, 128, -133])

        # 2-byte data
        data = d({u'isUnsigned': False, u'byteCount': 2},
                 [1, 2, -3, 32767, 1, -32768, -5])
        self.assertEqual(list(data), [1, 2, -3, 32768, -32773])

    def test_integer_packing_decoder_unsigned(self):
        """Test IntegerPacking decoder with unsigned data"""
        d = ihm.format_bcif._IntegerPackingDecoder()
        self.assertEqual(d._kind, u'IntegerPacking')

        # 1-byte data
        data = d({u'isUnsigned': True, u'byteCount': 1},
                 [1, 2, 3, 127, 1, 255, 1])
        self.assertEqual(list(data), [1, 2, 3, 127, 1, 256])

        # 2-byte data
        data = d({u'isUnsigned': True, u'byteCount': 2},
                 [1, 2, 3, 32767, 1, 65535, 5])
        self.assertEqual(list(data), [1, 2, 3, 32767, 1, 65540])

    def test_delta_decoder(self):
        """Test Delta decoder"""
        d = ihm.format_bcif._DeltaDecoder()
        self.assertEqual(d._kind, u'Delta')

        data = d({u'origin': 1000}, [0, 3, 2, 1])
        self.assertEqual(list(data), [1000, 1003, 1005, 1006])

    def test_run_length_decoder(self):
        """Test RunLength decoder"""
        d = ihm.format_bcif._RunLengthDecoder()
        self.assertEqual(d._kind, u'RunLength')

        data = d({}, [1, 3, 2, 1, 3, 2])
        self.assertEqual(list(data), [1, 1, 1, 2, 3, 3])

    def test_fixed_point_decoder(self):
        """Test FixedPoint decoder"""
        d = ihm.format_bcif._FixedPointDecoder()
        self.assertEqual(d._kind, u'FixedPoint')

        data = list(d({u'factor': 100}, [120, 123, 12]))
        self.assertEqual(len(data), 3)
        self.assertAlmostEqual(data[0], 1.20, delta=0.01)
        self.assertAlmostEqual(data[1], 1.23, delta=0.01)
        self.assertAlmostEqual(data[2], 0.12, delta=0.01)

    def test_decode(self):
        """Test _decode function"""
        data = b'\x01\x03\x02\x01\x03\x02'
        runlen = {u'kind': u'RunLength'}
        bytearr = {u'kind': u'ByteArray', u'type': ihm.format_bcif._Int8}
        data = ihm.format_bcif._decode(data, [runlen, bytearr])
        self.assertEqual(list(data), [1, 1, 1, 2, 3, 3])

    def _read_bcif(self, blocks, category_handlers,
                   unknown_category_handler=None,
                   unknown_keyword_handler=None):
        fh = _make_bcif_file(blocks)
        sys.modules['msgpack'] = MockMsgPack
        r = ihm.format_bcif.BinaryCifReader(fh, category_handlers,
                                            unknown_category_handler,
                                            unknown_keyword_handler)
        r.read_file()

    def test_category_case_insensitive(self):
        """Categories and keywords should be case insensitive"""
        cat1 = Category('_exptl', {'method': ['foo']})
        cat2 = Category('_Exptl', {'METHod': ['foo']})
        for cat in cat1, cat2:
            h = GenericHandler()
            self._read_bcif([Block([cat])], {'_exptl': h})
        self.assertEqual(h.data, [{'method': 'foo'}])

    def test_omitted_unknown(self):
        """Test handling of omitted/unknown data"""
        cat = Category('_foo',
                       {'var1': ['test1', '?', 'test2', None, 'test3']})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data,
                         [{'var1': 'test1'}, {'var1': '?'}, {'var1': 'test2'},
                          {}, {'var1': 'test3'}])

    def test_omitted_unknown_not_in_file_explicit(self):
        """Test explicit handling of omitted/unknown/not in file data"""
        cat = Category('_foo',
                       {'var1': ['test1', '?', 'test2', None, 'test3']})
        h = GenericHandler()
        h.omitted = 'OMIT'
        h.unknown = 'UNK'
        h.not_in_file = 'NOT'
        h._keys = ('var1', 'var2')
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data,
                         [{'var1': 'test1', 'var2': 'NOT'},
                          {'var1': 'UNK', 'var2': 'NOT'},
                          {'var1': 'test2', 'var2': 'NOT'},
                          {'var1': 'OMIT', 'var2': 'NOT'},
                          {'var1': 'test3', 'var2': 'NOT'}])

    def test_unknown_categories_ignored(self):
        """Check that unknown categories are just ignored"""
        cat1 = Category('_foo', {'var1': ['test1']})
        cat2 = Category('_bar', {'var2': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat1, cat2])], {'_foo': h})
        self.assertEqual(h.data, [{'var1': 'test1'}])

    def test_unknown_categories_handled(self):
        """Check that unknown categories are handled if requested"""
        class CatHandler(object):
            def __init__(self):
                self.warns = []

            def __call__(self, cat, line):
                self.warns.append((cat, line))

        ch = CatHandler()
        cat1 = Category('_foo', {'var1': ['test1']})
        cat2 = Category('_bar', {'var2': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat1, cat2])], {'_foo': h},
                        unknown_category_handler=ch)
        self.assertEqual(h.data, [{'var1': 'test1'}])
        self.assertEqual(ch.warns, [('_bar', None)])

    def test_unknown_keywords_ignored(self):
        """Check that unknown keywords are ignored"""
        cat = Category('_foo', {'var1': ['test1'], 'othervar': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'var1': 'test1'}])

    def test_unknown_keywords_handled(self):
        """Check that unknown keywords are handled if requested"""
        class KeyHandler(object):
            def __init__(self):
                self.warns = []

            def __call__(self, cat, key, line):
                self.warns.append((cat, key, line))

        kh = KeyHandler()
        cat = Category('_foo', {'var1': ['test1'], 'othervar': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h},
                        unknown_keyword_handler=kh)
        self.assertEqual(h.data, [{'var1': 'test1'}])
        self.assertEqual(kh.warns, [('_foo', 'othervar', None)])

    def test_multiple_data_blocks(self):
        """Test handling of multiple data blocks"""
        block1 = Block([Category('_foo',
                                 {'var1': ['test1'], 'var2': ['test2']})])
        block2 = Block([Category('_foo', {'var3': ['test3']})])
        fh = _make_bcif_file([block1, block2])

        h = GenericHandler()
        r = ihm.format_bcif.BinaryCifReader(fh, {'_foo': h})
        sys.modules['msgpack'] = MockMsgPack
        # Read first data block
        self.assertTrue(r.read_file())
        self.assertEqual(h.data, [{'var1': 'test1', 'var2': 'test2'}])

        # Read second data block
        h.data = []
        self.assertFalse(r.read_file())
        self.assertEqual(h.data, [{'var3': 'test3'}])

        # No more data blocks
        h.data = []
        self.assertFalse(r.read_file())
        self.assertEqual(h.data, [])

    def test_encoder(self):
        """Test _Encoder base class"""
        e = ihm.format_bcif._Encoder()
        e(None)  # noop

    def test_byte_array_encoder(self):
        """Test ByteArray encoder"""
        d = ihm.format_bcif._ByteArrayEncoder()

        # type 1 (signed char)
        data, encd = d([0, 1, -1])
        self.assertEqual(data, b'\x00\x01\xFF')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Int8})

        # type 2 (signed short)
        data, encd = d([256, -21503])
        self.assertEqual(data, b'\x00\x01\x01\xAC')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Int16})

        # type 3 (signed int)
        data, encd = d([-83951872])
        self.assertEqual(data, b'\x00\xff\xfe\xfa')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Int32})

        # type 4 (unsigned char)
        data, encd = d([0, 255])
        self.assertEqual(data, b'\x00\xFF')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Uint8})

        # type 5 (unsigned short)
        data, encd = d([256, 44033])
        self.assertEqual(data, b'\x00\x01\x01\xAC')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Uint16})

        # type 6 (unsigned int)
        data, encd = d([4278255872])
        self.assertEqual(data, b'\x00\x01\x01\xFF')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Uint32})

        # type 32 (32-bit float)
        data, encd = d([42.0])
        self.assertEqual(len(data), 4)
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Float32})

        # Too-large ints should cause an error
        self.assertRaises(TypeError, d, [2**34])
        self.assertRaises(TypeError, d, [-2**34])

    def test_delta_encoder(self):
        """Test Delta encoder"""
        d = ihm.format_bcif._DeltaEncoder()

        # too-small data is returned unchanged
        data = [0, 1, -1]
        encdata, encdict = d(data)
        self.assertEqual(data, encdata)
        self.assertIsNone(encdict)

        # large data is encoded
        data = [0, 1, -1] + [-1] * 40
        encdata, encdict = d(data)
        self.assertEqual(encdata, [0, 1, -2] + [0] * 40)
        self.assertEqual(encdict, {'origin': 0, 'kind': 'Delta',
                                   'srcType': ihm.format_bcif._Int8})

    def test_run_length_encoder(self):
        """Test RunLength encoder"""
        d = ihm.format_bcif._RunLengthEncoder()

        # too-small data is returned unchanged
        data = [0, 1, -1]
        encdata, encdict = d(data)
        self.assertEqual(data, encdata)
        self.assertIsNone(encdict)

        # large data that can't be compressed is returned unchanged
        data = list(range(50))
        encdata, encdict = d(data)
        self.assertEqual(data, encdata)
        self.assertIsNone(encdict)

        # large data that can be compressed
        data = [0] * 30 + [1] * 40
        encdata, encdict = d(data)
        self.assertEqual(encdata, [0, 30, 1, 40])
        self.assertEqual(encdict, {'kind': 'RunLength', 'srcSize': 70,
                                   'srcType': ihm.format_bcif._Uint8})

    def test_encode(self):
        """Test _encode function"""
        data = [1, 1, 1, 2, 3, 3]
        encoders = [ihm.format_bcif._ByteArrayEncoder()]
        encdata, encds = ihm.format_bcif._encode(data, encoders)
        self.assertEqual(encdata, b'\x01\x01\x01\x02\x03\x03')
        self.assertEqual(encds, [{'kind': 'ByteArray',
                                  'type': ihm.format_bcif._Uint8}])

        # DeltaEncoder will be a noop here since data is small
        encoders = [ihm.format_bcif._DeltaEncoder(),
                    ihm.format_bcif._ByteArrayEncoder()]
        encdata, encds = ihm.format_bcif._encode(data, encoders)
        self.assertEqual(encdata, b'\x01\x01\x01\x02\x03\x03')
        self.assertEqual(encds, [{'kind': 'ByteArray',
                                  'type': ihm.format_bcif._Uint8}])

    def test_mask_type_no_mask(self):
        """Test get_mask_and_type with no mask"""
        data = [1, 2, 3, 4]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertIsNone(mask)
        self.assertEqual(typ, int)

    def test_mask_type_masked_int(self):
        """Test get_mask_and_type with masked int data"""
        data = [1, 2, 3, None, ihm.unknown, 4]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 0, 0, 1, 2, 0])
        self.assertEqual(typ, int)

    def test_mask_type_masked_long(self):
        """Test get_mask_and_type with masked long data"""
        if sys.version_info[0] < 3:
            # long type is only in Python 2
            # Use long(x) rather than xL since the latter will cause a syntax
            # error in Python 3
            data = [long(1), long(2), long(3),    # noqa: F821
                    None, ihm.unknown, long(4)]   # noqa: F821
            mask, typ = ihm.format_bcif._get_mask_and_type(data)
            self.assertEqual(mask, [0, 0, 0, 1, 2, 0])
            self.assertEqual(typ, int)

    def test_mask_type_masked_float(self):
        """Test get_mask_and_type with masked float data"""
        data = [1.0, 2.0, 3.0, None, ihm.unknown, 4.0]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 0, 0, 1, 2, 0])
        self.assertEqual(typ, float)

    def test_mask_type_masked_numpy_float(self):
        """Test get_mask_and_type with masked numpy float data"""
        try:
            import numpy
        except ImportError:
            self.skipTest("this test requires numpy")
        data = [numpy.float64(4.2), None, ihm.unknown]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 1, 2])
        self.assertEqual(typ, float)

    def test_mask_type_masked_str(self):
        """Test get_mask_and_type with masked str data"""
        # Literal . and ? should not be masked
        data = ['a', 'b', None, ihm.unknown, 'c', '.', '?']
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 0, 1, 2, 0, 0, 0])
        self.assertEqual(typ, str)

    def test_mask_type_mix_int_float(self):
        """Test get_mask_and_type with a mix of int and float data"""
        data = [1, 2, 3, 4.0]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertIsNone(mask)
        self.assertEqual(typ, float)  # int/float is coerced to float

    def test_mask_type_mix_int_float_str(self):
        """Test get_mask_and_type with a mix of int/float/str data"""
        data = [1, 2, 3, 4.0, 'foo']
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertIsNone(mask)
        self.assertEqual(typ, str)  # int/float/str is coerced to str

    def test_mask_type_bad_type(self):
        """Test get_mask_and_type with unknown type data"""
        class MockObject(object):
            pass
        data = [MockObject()]
        self.assertRaises(ValueError, ihm.format_bcif._get_mask_and_type, data)

    def test_masked_encoder(self):
        """Test MaskedEncoder base class"""
        e = ihm.format_bcif._MaskedEncoder()
        e(None, None)  # noop

    def test_string_array_encoder_no_mask(self):
        """Test StringArray encoder with no mask"""
        d = ihm.format_bcif._StringArrayMaskedEncoder()
        indices, encs = d(['a', 'AB', 'a'], None)
        self.assertEqual(indices, b'\x00\x01\x00')
        enc, = encs
        self.assertEqual(enc['dataEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Uint8}])
        self.assertEqual(enc['offsetEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Uint8}])
        self.assertEqual(enc['offsets'], b'\x00\x01\x03')
        self.assertEqual(enc['stringData'], 'aAB')

    def test_string_array_encoder_mask(self):
        """Test StringArray encoder with mask"""
        d = ihm.format_bcif._StringArrayMaskedEncoder()
        # True should be mapped to 'YES'; int 3 to str '3'
        # Unmasked literal . and ? should be kept as-is
        indices, encs = d(['a', 'AB', True, ihm.unknown, None, 'a', 3,
                           '.', '?'],
                          [0, 0, 0, 2, 1, 0, 0, 0, 0])
        # \xff is -1 (masked value) as a signed char (Int8)
        self.assertEqual(indices, b'\x00\x01\x02\xff\xff\x00\x03\x04\x05')
        enc, = encs
        self.assertEqual(enc['dataEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Int8}])
        self.assertEqual(enc['offsetEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Uint8}])
        self.assertEqual(enc['offsets'], b'\x00\x01\x03\x06\x07\x08\t')
        self.assertEqual(enc['stringData'], 'aABYES3.?')

    def test_int_array_encoder_no_mask(self):
        """Test IntArray encoder with no mask"""
        d = ihm.format_bcif._IntArrayMaskedEncoder()
        data, encs = d([5, 7, 8], None)
        self.assertEqual(data, b'\x05\x07\x08')
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Uint8}])

    def test_int_array_encoder_mask(self):
        """Test IntArray encoder with mask"""
        d = ihm.format_bcif._IntArrayMaskedEncoder()
        data, encs = d([5, 7, '?', 8, None], [0, 0, 2, 0, 1])
        # \xff is -1 (masked value) as a signed char (Int8)
        self.assertEqual(data, b'\x05\x07\xff\x08\xff')
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Int8}])

    def test_float_array_encoder_no_mask(self):
        """Test FloatArray encoder with no mask"""
        d = ihm.format_bcif._FloatArrayMaskedEncoder()
        # int data should be coerced to float
        data, encs = d([5.0, 7.0, 8.0, 4], None)
        self.assertEqual(len(data), 4 * 4)
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Float32}])

    def test_float_array_encoder_mask(self):
        """Test FloatArray encoder with mask"""
        d = ihm.format_bcif._FloatArrayMaskedEncoder()
        data, encs = d([5., 7., '?', 8., None], [0, 0, 2, 0, 1])
        self.assertEqual(len(data), 5 * 4)
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Float32}])

    def test_category(self):
        """Test CategoryWriter class"""
        fh = MockFh()
        sys.modules['msgpack'] = MockMsgPack
        writer = ihm.format_bcif.BinaryCifWriter(fh)
        writer.start_block('ihm')
        with writer.category('foo') as loc:
            loc.write(bar='baz')
        writer.flush()
        block, = fh.data['dataBlocks']
        category, = block['categories']
        column, = category['columns']
        self.assertEqual(block['header'], 'ihm')
        self.assertEqual(category['name'], 'foo')
        self.assertEqual(category['rowCount'], 1)
        self.assertEqual(column['name'], 'bar')
        self.assertEqual(column['data']['encoding'][0]['stringData'],
                         'baz')

    def test_empty_loop(self):
        """Test LoopWriter class with no values"""
        fh = MockFh()
        sys.modules['msgpack'] = MockMsgPack
        writer = ihm.format_bcif.BinaryCifWriter(fh)
        writer.start_block('ihm')
        with writer.loop('foo', ["bar", "baz"]):
            pass
        writer.flush()
        self.assertEqual(fh.data['dataBlocks'][0]['categories'], [])

    def test_loop(self):
        """Test LoopWriter class"""
        fh = MockFh()
        sys.modules['msgpack'] = MockMsgPack
        writer = ihm.format_bcif.BinaryCifWriter(fh)
        writer.start_block('ihm')
        with writer.loop('foo', ["bar", "baz"]) as lp:
            lp.write(bar='x')
            lp.write(bar=None, baz='z')
            lp.write(bar=ihm.unknown, baz='z')
            lp.write(bar='.', baz='z')
            lp.write(bar='?', baz='z')
            lp.write(baz='y')
        writer.flush()
        block, = fh.data['dataBlocks']
        category, = block['categories']
        self.assertEqual(category['name'], 'foo')
        self.assertEqual(category['rowCount'], 6)
        cols = sorted(category['columns'], key=lambda x: x['name'])
        self.assertEqual(len(cols), 2)
        # Check mask for bar column; literal . and ? should not be masked (=0)
        self.assertEqual(cols[0]['mask']['data'],
                         b'\x00\x01\x02\x00\x00\x01')


if __name__ == '__main__':
    unittest.main()



# This example demonstrates the use of the Python IHM library to generate
# an mmCIF file for a very simple integrative docking study. Two subunits,
# A and B, each of which is fitted against small angle X-ray (SAXS) data, are
# docked together into a complex, AB, which is fitted against an electron
# microscopy density map.

import ihm
import ihm.location
import ihm.dataset
import ihm.representation
import ihm.restraint
import ihm.protocol
import ihm.model
import ihm.dumper

# First, we create a system, which contains everything we know about the
# modeling. A single mmCIF file can contain multiple Systems, but in most
# cases we use just one:
system = ihm.System()

# Next, we describe the input data we used, using dataset classes.
# Each source of data has a location, such as a file on disk or a database
# entry, and a type. In this example we used EM density data, which we'll
# say lives in the EMDB database:
loc = ihm.location.EMDBLocation('EMDB-1234')
em_dataset = ihm.dataset.EMDensityDataset(loc)
# We also used two SAXS profiles, which we'll say live in SASBDB:
saxsA_dataset = ihm.dataset.SASDataset(ihm.location.SASBDBLocation('SASDB123'))
saxsB_dataset = ihm.dataset.SASDataset(ihm.location.SASBDBLocation('SASDB456'))

# Where datasets are derived from some other data, it is helpful to also point
# back to that primary data. In this case, let's say the EM density was
# derived from a set of EM micrographs, deposited in the EMPIAR database:
m = ihm.dataset.EMMicrographsDataset(ihm.location.EMPIARLocation('EMPIAR-123'))
em_dataset.parents.append(m)

# Next, define the entities for each unique sequence in the system
# (here represented as polyalanines):
entityA = ihm.Entity('AAA', description='Subunit A')
entityB = ihm.Entity('AAAAAA', description='Subunit B')
system.entities.extend((entityA, entityB))

# Next, we define asymmetric units for everything we modeled.
# These roughly correspond to chains in a traditional PDB file. Multiple
# asymmetric units may map to the same entity (for example if there are
# several copies of a given protein). Parts of the system that were seen in
# an experiment but were not modeled are represented as entities to which no
# asymmetric units map.
asymA = ihm.AsymUnit(entityA, details='Subunit A')
asymB = ihm.AsymUnit(entityB, details='Subunit B')
system.asym_units.extend((asymA, asymB))

# Next, we group asymmetric units (and/or entities) into assemblies.
# Here, we'll define an assembly of everything that we modeled, plus
# two subassemblies (of the subunits) that the SAXS data applies to:
modeled_assembly = ihm.Assembly((asymA, asymB), name='Modeled assembly')
assemblyA = ihm.Assembly((asymA,), name='Subunit A')
assemblyB = ihm.Assembly((asymB,), name='Subunit B')

# Define how the system was represented. Multiple representations of the
# system are possible, and can overlap. Here we'll say we represent A
# atomically as a rigid body and B as 3 flexible coarse-grained spheres:
rep = ihm.representation.Representation(
    [ihm.representation.AtomicSegment(asymA, rigid=True),
     ihm.representation.FeatureSegment(asymB, rigid=False,
                                       primitive='sphere', count=3)])

# Set up restraints on the system. First, two on the subunits that use
# the SAXS data; we'll say we used the FoXS software to do this fit:
saxsA_rsr = ihm.restraint.SASRestraint(
    dataset=saxsA_dataset, assembly=assemblyA,
    fitting_method='FoXS', fitting_atom_type='Heavy atoms')
saxsB_rsr = ihm.restraint.SASRestraint(
    dataset=saxsB_dataset, assembly=assemblyB,
    fitting_method='FoXS', fitting_atom_type='Heavy atoms')
system.restraints.extend((saxsA_rsr, saxsB_rsr))
# Next, the EM restraint applied to the entire system:
em_rsr = ihm.restraint.EM3DRestraint(
    dataset=em_dataset, assembly=modeled_assembly)
system.restraints.append(em_rsr)

# Now we add information about how the modeling was done by defining one
# or more protocols. Here we'll say we did simple Monte Carlo on the entire
# system using all of the experimental data:
all_datasets = ihm.dataset.DatasetGroup((em_dataset, saxsA_dataset,
                                         saxsB_dataset))
protocol = ihm.protocol.Protocol(name='Modeling')
protocol.steps.append(ihm.protocol.Step(
    assembly=modeled_assembly, dataset_group=all_datasets,
    method='Monte Carlo', name='Production sampling',
    num_models_begin=0, num_models_end=1000, multi_scale=True))

# Finally we can add coordinates for the deposited models. Typically these
# will be stored in our own software's data structures somewhere (for this
# example in simple lists 'atoms' and 'spheres'):
atoms = [('A', 1, 'C', 'CA', 1., 2., 3.),
         ('A', 2, 'C', 'CA', 4., 5., 6.),
         ('A', 3, 'C', 'CA', 7., 8., 9.)]
spheres = [('B', 1, 2, 1., 2., 3., 1.2),
           ('B', 3, 4, 4., 5., 6., 1.2),
           ('B', 5, 6, 7., 8., 9., 1.2)]


# Rather than storing another copy of the coordinates in the IHM library
# (which could use a lot of memory), we need to provide a mechanism to
# translate them into the IHM data model. We do this straighforwardly by
# subclassing the IHM Model class and overriding the get_atoms
# and get_spheres methods:
class MyModel(ihm.model.Model):
    # Map our asym unit names A and B to IHM asym_unit objects:
    asym_unit_map = {'A': asymA, 'B': asymB}

    def get_atoms(self):
        for asym, seq_id, type_symbol, atom_id, x, y, z in atoms:
            yield ihm.model.Atom(asym_unit=self.asym_unit_map[asym],
                                 type_symbol=type_symbol, seq_id=seq_id,
                                 atom_id=atom_id, x=x, y=y, z=z)

    def get_spheres(self):
        for asym, seq_id_start, seq_id_end, x, y, z, radius in spheres:
            yield ihm.model.Sphere(asym_unit=self.asym_unit_map[asym],
                                   seq_id_range=(seq_id_start, seq_id_end),
                                   x=x, y=y, z=z, radius=radius)


model = MyModel(assembly=modeled_assembly, protocol=protocol,
                representation=rep, name='Best scoring model')

# Note that the model was scored against all three restraints
saxsA_rsr.fits[model] = ihm.restraint.SASRestraintFit(chi_value=1.4)
saxsB_rsr.fits[model] = ihm.restraint.SASRestraintFit(chi_value=2.1)
em_rsr.fits[model] = ihm.restraint.EM3DRestraintFit(
    cross_correlation_coefficient=0.9)

# Similar models can be grouped together. Here we only have a single model
# in the group
model_group = ihm.model.ModelGroup([model], name='All models')

# Groups are then placed into states, which can in turn be grouped. In this
# case we have only a single state:
state = ihm.model.State([model_group])
system.state_groups.append(ihm.model.StateGroup([state]))

# Once the system is complete, we can write it out to an mmCIF file:
with open('output.cif', 'w') as fh:
    ihm.dumper.write(fh, [system])



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.cross_linkers


class Tests(unittest.TestCase):

    def test_cross_linkers(self):
        """Test cross_linkers module"""
        d = ihm.cross_linkers.dss
        self.assertEqual(d.auth_name, 'DSS')
        self.assertEqual(d.smiles,
                         'C1CC(=O)N(C1=O)OC(=O)CCCCCCC(=O)ON2C(=O)CCC2=O')


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.dataset
import ihm.location
import ihm.geometry


def _make_test_file(fname):
    with open(fname, 'w') as fh:
        fh.write('contents')


class Tests(unittest.TestCase):

    def test_dataset(self):
        """Test Dataset base class"""
        loc = ihm.location.PDBLocation('1abc', version='foo', details='bar')
        d = ihm.dataset.Dataset(loc)
        self.assertIsNone(d.details)
        self.assertEqual(len(d.parents), 0)

        l2 = ihm.location.PDBLocation('1xyz', version='foo', details='bar')
        d2 = ihm.dataset.Dataset(l2, details='foo')
        self.assertEqual(d2.details, 'foo')
        d.parents.append(d2)
        self.assertEqual(len(d.parents), 1)
        self.assertNotEqual(d, d2)

        l3 = ihm.location.PDBLocation('1cde', version='foo', details='bar')
        d3 = ihm.dataset.Dataset(l3, details='bar')
        t = ihm.geometry.Transformation(
            rot_matrix=[[-0.64, 0.09, 0.77], [0.76, -0.12, 0.64],
                        [0.15, 0.99, 0.01]],
            tr_vector=[1., 2., 3.])
        td = ihm.dataset.TransformedDataset(d3, transform=t)
        d.parents.append(td)
        self.assertEqual(len(d.parents), 2)

    def test_add_primary_no_parents(self):
        """Test add_primary() method, no parents"""
        l1 = ihm.location.PDBLocation('1abc', version='foo', details='bar')
        d1 = ihm.dataset.Dataset(l1)

        l2 = ihm.location.PDBLocation('1xyz', version='foo', details='bar')
        d2 = ihm.dataset.Dataset(l2)

        d1.add_primary(d2)
        self.assertEqual(d1.parents, [d2])

    def test_add_primary_one_parent(self):
        """Test add_primary() method, one parent"""
        l1 = ihm.location.PDBLocation('1abc', version='foo', details='bar')
        d1 = ihm.dataset.Dataset(l1)

        l2 = ihm.location.PDBLocation('1xyz', version='foo', details='bar')
        d2 = ihm.dataset.Dataset(l2)

        l3 = ihm.location.PDBLocation('2def', version='foo', details='bar')
        d3 = ihm.dataset.Dataset(l3)

        d1.parents.append(d2)

        d1.add_primary(d3)
        self.assertEqual(d1.parents, [d2])
        self.assertEqual(d2.parents, [d3])

    def test_add_primary_two_parents(self):
        """Test add_primary() method, two parents"""
        l1 = ihm.location.PDBLocation('1abc', version='foo', details='bar')
        d1 = ihm.dataset.Dataset(l1)

        l2 = ihm.location.PDBLocation('1xyz', version='foo', details='bar')
        d2 = ihm.dataset.Dataset(l2)

        l3 = ihm.location.PDBLocation('2def', version='foo', details='bar')
        d3 = ihm.dataset.Dataset(l3)

        _ = ihm.location.PDBLocation('2ghi', version='foo', details='bar')
        d4 = ihm.dataset.Dataset(l3)

        d1.parents.extend((d2, d3))

        self.assertRaises(ValueError, d1.add_primary, d4)

    def test_cxms_dataset(self):
        """Test CXMSDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.CXMSDataset(loc)
        self.assertEqual(d.data_type, 'CX-MS data')

    def test_hdx_dataset(self):
        """Test HDXDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.HDXDataset(loc)
        self.assertEqual(d.data_type, 'H/D exchange data')

    def test_mass_spec_dataset(self):
        """Test MassSpecDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.MassSpecDataset(loc)
        self.assertEqual(d.data_type, 'Mass Spectrometry data')

    def test_em_density_dataset(self):
        """Test EMDensityDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.EMDensityDataset(loc)
        self.assertEqual(d.data_type, '3DEM volume')

    def test_pdb_dataset(self):
        """Test PDBDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.PDBDataset(loc)
        self.assertEqual(d.data_type, 'Experimental model')

    def test_comp_model_dataset(self):
        """Test ComparativeModelDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.ComparativeModelDataset(loc)
        self.assertEqual(d.data_type, 'Comparative model')

    def test_int_model_dataset(self):
        """Test IntegrativeModelDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.IntegrativeModelDataset(loc)
        self.assertEqual(d.data_type, 'Integrative model')

    def test_de_novo_model_dataset(self):
        """Test DeNovoModelDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.DeNovoModelDataset(loc)
        self.assertEqual(d.data_type, 'De Novo model')

    def test_nmr_dataset(self):
        """Test NMRDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.NMRDataset(loc)
        self.assertEqual(d.data_type, 'NMR data')

    def test_mutagenesis_dataset(self):
        """Test MutagenesisDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.MutagenesisDataset(loc)
        self.assertEqual(d.data_type, 'Mutagenesis data')

    def test_em2d_class_dataset(self):
        """Test EM2DClassDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.EM2DClassDataset(loc)
        self.assertEqual(d.data_type, '2DEM class average')

    def test_em_micrographs_dataset(self):
        """Test EMMicrographsDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.EMMicrographsDataset(loc)
        self.assertEqual(d.data_type, 'EM raw micrographs')

    def test_sas_dataset(self):
        """Test SASDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.SASDataset(loc)
        self.assertEqual(d.data_type, 'SAS data')

    def test_fret_dataset(self):
        """Test FRETDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.FRETDataset(loc)
        self.assertEqual(d.data_type, 'Single molecule FRET data')

    def test_y2h_dataset(self):
        """Test YeastTwoHybridDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.YeastTwoHybridDataset(loc)
        self.assertEqual(d.data_type, 'Yeast two-hybrid screening data')

    def test_genetic_dataset(self):
        """Test GeneticInteractionsDataset"""
        loc = ihm.location.FileLocation(repo='mydoi', path='a')
        d = ihm.dataset.GeneticInteractionsDataset(loc)
        self.assertEqual(
            d.data_type, 'Quantitative measurements of genetic interactions')

    def test_duplicate_datasets_details(self):
        """Datasets with differing details should be considered duplicates"""
        with utils.temporary_directory() as tmpdir:
            fname = os.path.join(tmpdir, 'test.pdb')
            _make_test_file(fname)
            l1 = ihm.location.InputFileLocation(fname, details='test details')
            _ = ihm.dataset.PDBDataset(l1)

            l2 = ihm.location.InputFileLocation(fname, details='other details')
            d2 = ihm.dataset.PDBDataset(l2)
            self.assertEqual(l1, l2)
            d3 = ihm.dataset.PDBDataset(l2, details='other dataset details')
            self.assertEqual(d2, d3)

    def test_duplicate_locations(self):
        """Datasets with same location should be considered duplicates"""
        with utils.temporary_directory() as tmpdir:
            fname1 = os.path.join(tmpdir, 'test1.pdb')
            fname2 = os.path.join(tmpdir, 'test2.pdb')
            _make_test_file(fname1)
            _make_test_file(fname2)
            loc1 = ihm.location.InputFileLocation(fname1)
            loc2 = ihm.location.InputFileLocation(fname2)

            # Identical datasets in the same location aren't duplicated
            pdb1 = ihm.dataset.PDBDataset(loc1)
            pdb2 = ihm.dataset.PDBDataset(loc1)
            self.assertEqual(pdb1, pdb2)

            # Datasets in different locations are OK
            pdb1 = ihm.dataset.PDBDataset(loc1)
            pdb2 = ihm.dataset.PDBDataset(loc2)
            self.assertNotEqual(pdb1, pdb2)


if __name__ == '__main__':
    unittest.main()
